"""Module defining event handler types."""

from typing import Callable

from aim.sdk.callbacks.helpers import (
    add_handler_event_name,
    check_handler_method_params,
)


class On:
    def __getattr__(self, event_name: str) -> Callable:
        def mark_as_event_handler(method: Callable):
            check_handler_method_params(method)
            add_handler_event_name(method, event_name)
            return method

        return mark_as_event_handler


on = On()


__all__ = ['on']



import json
import time
import queue
import logging

from threading import Thread
from pathlib import Path
from cachetools.func import ttl_cache
from typing import Dict, Optional, Generic, TypeVar
from abc import abstractmethod
from collections import OrderedDict

from aim.sdk.repo import Repo
from aim.sdk.run import Run
from aim.storage.locking import AutoFileLock
from aim.ext.notifier import get_config, get_notifier, Notifier, NotificationSendError
from aim.ext.notifier.utils import get_working_directory
from aim.ext.cleanup import AutoClean


logger = logging.getLogger(__name__)

GRACE_PERIOD = 100  # seconds

T = TypeVar('T')


class RepetitionCounter:
    def __init__(self, obj):
        self.obj = obj
        self.step = 0
        self.count = 0

    def update(self, step):
        self.count += 1
        self.step = max(step, self.step)


class RankedSet(Generic[T]):
    def __init__(self):
        self.events: Dict[str, T] = {}

    def get(self, obj_idx: str) -> T:
        return self.events[obj_idx]

    def add(self, event: T) -> bool:
        object_idx = event.obj_idx
        if object_idx not in self.events:
            self.events[object_idx] = event
            return True
        elif self.events[object_idx].idx < event.idx:
            self.events[object_idx] = event
            return True
        return False


class RunVariable:
    """Representation of Run object with run.hash only. To be replaced with real Run in the future."""
    def __init__(self, run_hash: str):
        self.hash = run_hash


class Event:
    def __init__(self, status_event_encoded: str):
        obj_idx, idx, event_type, epoch_time, next_event_in = status_event_encoded.split('-')
        self.idx: int = int(idx)
        self.obj_idx: str = obj_idx

        self.event_type = event_type
        self.next_event_in: int = int(next_event_in)
        self.epoch_time: float = float(epoch_time)
        self.detected_epoch_time: float = time.time()


EventSet = RankedSet[Event]


class Notification:
    def __init__(self, *,
                 obj_idx: Optional[str] = None,
                 rank: Optional[int] = None,
                 message: Optional[str] = None):
        self.rank = rank
        self.obj_idx = obj_idx
        self.message = message

    @abstractmethod
    def is_sent(self):
        ...

    @abstractmethod
    def update_last_sent(self):
        ...

    @abstractmethod
    def get_msg_details(self):
        ...


class StatusNotification(Notification):
    _notifications_cache_path: Path = None
    notifications_cache: Dict[str, int] = {}

    @classmethod
    def recover_notifications_cache(cls, cache_path: Path):
        cls._notifications_cache_path = cache_path
        with cls._notifications_cache_path.open() as cache_fh:
            try:
                cls.notifications_cache = json.load(cache_fh)
            except json.decoder.JSONDecodeError:
                cls.notifications_cache = {}

    def get_msg_details(self):
        return {'run': RunVariable(self.obj_idx)}

    def is_sent(self) -> bool:
        last_event_idx = self.notifications_cache.get(self.obj_idx, -1)
        if self.rank is None or last_event_idx < self.rank:
            return False
        elif last_event_idx == self.rank:
            return True
        else:
            logger.warning(f'New event id {self.rank} for object \'{self.obj_idx}\' '
                           f'is less than last reported event id {last_event_idx}.')
            return True

    def update_last_sent(self):
        if self.rank is not None:
            self.notifications_cache[self.obj_idx] = self.rank
            with self._notifications_cache_path.open(mode='w') as notifications_fh:
                json.dump(self.notifications_cache, notifications_fh)


class LogNotification(Notification):
    _repo: Repo = None

    @classmethod
    def set_repo(cls, repo: Repo):
        cls._repo = repo

    def get_msg_details(self):
        return {}

    def is_sent(self) -> bool:
        run = self._repo.structured_db.find_run(self.obj_idx)
        return self.rank < run.info.last_notification_index

    def update_last_sent(self):
        run = self._repo.structured_db.find_run(self.obj_idx)
        run.info.last_notification_index = self.rank


class WorkerThread(Thread):
    def __init__(self, func, **kwargs):
        self.shutdown = False
        self.func = func
        super().__init__(target=self.worker, **kwargs)

    def worker(self):
        while True:
            if self.shutdown:
                break
            self.func()

    def stop(self):
        logger.debug('Stopping worker thread...')
        self.shutdown = True
        self.join()
        logger.debug('Worker thread stopped.')


class NotificationQueue(object):
    def __init__(self, notifier: Notifier):
        self._stopped = False
        self._notifier = notifier

        self._queue = queue.Queue()
        self._notifier_thread = Thread(target=self.listen, daemon=True)
        self._notifier_thread.start()

    def add_notification(self, notification: Notification):
        if not notification.is_sent():
            self._queue.put(notification)

    def stop(self):
        logger.debug('Processing remaining notifications...')
        self._queue.join()
        logger.debug('Notifications queue is empty.')
        logger.debug('Stopping worker thread...')
        self._stopped = True
        self._notifier_thread.join()
        logger.debug('Worker thread stopped.')

    def listen(self):
        while True:
            if self._stopped:
                break
            try:
                notification = self._queue.get(timeout=1)
                if notification.is_sent():
                    logger.debug(f'Notification for object \'{notification.obj_idx}\' '
                                 f'with event ID {notification.event_idx} has already been sent. Skipping.')
                else:
                    details = notification.get_msg_details()
                    try:
                        self._notifier.notify(notification.message, **details)
                        notification.update_last_sent()
                    except NotificationSendError as e:
                        logger.error(f'Failed to send notification. Reason: {e}.')
                self._queue.task_done()
            except queue.Empty:
                continue


class RunStatusWatcherAutoClean(AutoClean['RunStatusWatcher']):
    PRIORITY = 30

    def __init__(self, instance: 'RunStatusWatcher') -> None:
        super().__init__(instance)
        self.is_background = instance.background
        self.watcher_thread = instance.watcher_thread
        self.notifications_queue = instance.notifications_queue
        self.lock = instance.lock

    def _close(self) -> None:
        if self.is_background:
            assert self.watcher_thread is not None
            self.watcher_thread.stop()
        self.notifications_queue.stop()
        self.lock.release()


class RunStatusWatcher:
    message_templates = {
        'finished': 'Run \'{run.hash}\' finished without error.',
        'starting': 'Run \'{run.hash}\' started.'
    }

    def __init__(self, repo: Repo, background: bool = False):
        self.repo_path = Path(repo.path)
        self.repo = repo
        self.background = background

        self._resources = None
        self._log_lvl_threshold = None
        self.initialized = False

        work_dir = get_working_directory(self.repo_path)
        self.lock = AutoFileLock(work_dir / 'watcher.lock', timeout=0)
        try:
            self.lock.acquire()
        except TimeoutError:
            logger.error(f'Cannot start Run status watcher for \'{self.repo_path}\'. Failed to acquire lock.')
            return

        self._status_watch_dir: Path = self.repo_path / 'check_ins'
        self._status_watch_dir.mkdir(exist_ok=True)

        self._notifications_cache_path: Path = work_dir / 'last_run_notifications'
        self._notifications_cache_path.touch(exist_ok=True)
        self._status_events = EventSet()
        self._log_events = EventSet()

        self.notifier = get_notifier(self.repo_path)
        self.watcher_thread = WorkerThread(self.check_for_new_events, daemon=True) if not background else None
        self.notifications_queue = NotificationQueue(self.notifier)

        StatusNotification.recover_notifications_cache(self._notifications_cache_path)
        LogNotification.set_repo(self.repo)

        self.initialized = True
        self._resources = RunStatusWatcherAutoClean(self)

    @property
    @ttl_cache(maxsize=None, ttl=10)
    def log_level_threshold(self):
        lvl = get_config(self.repo_path).log_level
        if self._log_lvl_threshold is not None:
            if self._log_lvl_threshold != lvl:
                logger.warning(f'Log Notifications level changed '
                               f'from \'{logging.getLevelName(self._log_lvl_threshold)}\' '
                               f'to \'{logging.getLevelName(lvl)}\'.')
                self._log_lvl_threshold = lvl
        else:
            logger.warning(f'Running with Log Notifications level \'{logging.getLevelName(lvl)}\'')
            self._log_lvl_threshold = lvl
        return lvl

    def start_watcher(self):
        if not self.initialized:
            return
        logger.info('Starting watcher...')
        notification = StatusNotification(message=f'Watcher is running for repo \'{self.repo_path}\'')
        self.notifications_queue.add_notification(notification)
        if self.background:
            self.watcher_thread.start()
        else:
            self.run_forever()

    def run_forever(self):
        while True:
            self.check_for_new_events()
            time.sleep(1)

    def check_for_new_events(self):
        status_events = self.poll_status_events()
        for new_event in status_events.events.values():
            if not self._status_events.add(new_event):
                event = self._status_events.get(new_event.obj_idx)
                if event.next_event_in == 0:  # wait for next check-in for infinite time
                    continue
                epoch_now = time.time()
                failure = (event.next_event_in + GRACE_PERIOD < epoch_now - event.detected_epoch_time)
                if failure:
                    notification = StatusNotification(obj_idx=event.obj_idx, rank=event.idx)
                    self.notifications_queue.add_notification(notification)
            elif new_event.event_type in self.message_templates:
                message = self.message_templates[new_event.event_type]
                notification = StatusNotification(obj_idx=new_event.obj_idx, rank=new_event.idx, message=message)
                self.notifications_queue.add_notification(notification)

        self.repo.persistent_pool.clear()

        log_events = self.poll_log_record_events()
        for new_event in log_events.events.values():
            if self._log_events.add(new_event):
                run_hash = new_event.obj_idx
                run = Run(run_hash, repo=self.repo, read_only=True)
                run_info = run.props.info
                log_records_seq = run.get_log_records()
                if log_records_seq:
                    last_log_index = log_records_seq.last_step()
                    last_notified_index = run_info.last_notification_index
                    if last_log_index > last_notified_index:
                        data = log_records_seq.data.range(last_notified_index + 1, last_log_index + 1)
                        for rec_hash, counter in self._processed_log_records(data).items():
                            assert counter.count > 0
                            if counter.count == 1:
                                message = counter.obj.message
                            else:
                                message = f'{counter.obj.message} ({counter.count - 1} more messages logged)'
                            notification = LogNotification(obj_idx=run_hash, rank=counter.step, message=message)
                            self.notifications_queue.add_notification(notification)

    def poll_status_events(self) -> EventSet:
        return self._poll(event_types=('finished', 'starting', 'check_in'))

    def poll_log_record_events(self) -> EventSet:
        return self._poll(event_types=('new_logs',))

    def _poll(self, event_types) -> EventSet:
        events = EventSet()
        for event_type in event_types:
            for check_in_file_path in sorted(self._status_watch_dir.glob(f'*-*-{event_type}-*-*')):
                events.add(Event(check_in_file_path.name))
        return events

    def _processed_log_records(self, log_records_data) -> Dict:
        log_level = self.log_level_threshold

        steps, log_records = log_records_data.view('val').items_list()
        log_records = log_records[0]

        log_info_map = OrderedDict()
        for step, log_record in zip(steps, log_records):
            if log_record.level >= log_level:
                log_info_map.setdefault(hash(log_record), RepetitionCounter(log_record)).update(step)
        return log_info_map



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
# temporary workaround for M1 build
try:
    import grpc
except ImportError:
    grpc = None

import aim.ext.transport.remote_router_pb2 as remote__router__pb2


class RemoteRouterServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.client_heartbeat = channel.unary_unary(
                '/RemoteRouterService/client_heartbeat',
                request_serializer=remote__router__pb2.HeartbeatRequest.SerializeToString,
                response_deserializer=remote__router__pb2.HeartbeatResponse.FromString,
                )
        self.get_version = channel.unary_unary(
                '/RemoteRouterService/get_version',
                request_serializer=remote__router__pb2.VersionRequest.SerializeToString,
                response_deserializer=remote__router__pb2.VersionResponse.FromString,
                )
        self.connect = channel.unary_unary(
                '/RemoteRouterService/connect',
                request_serializer=remote__router__pb2.ConnectRequest.SerializeToString,
                response_deserializer=remote__router__pb2.ConnectResponse.FromString,
                )
        self.reconnect = channel.unary_unary(
                '/RemoteRouterService/reconnect',
                request_serializer=remote__router__pb2.ReconnectRequest.SerializeToString,
                response_deserializer=remote__router__pb2.ReconnectResponse.FromString,
                )
        self.disconnect = channel.unary_unary(
                '/RemoteRouterService/disconnect',
                request_serializer=remote__router__pb2.DisconnectRequest.SerializeToString,
                response_deserializer=remote__router__pb2.DisconnectResponse.FromString,
                )


class RemoteRouterServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def client_heartbeat(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def get_version(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def connect(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def reconnect(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def disconnect(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_RemoteRouterServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'client_heartbeat': grpc.unary_unary_rpc_method_handler(
                    servicer.client_heartbeat,
                    request_deserializer=remote__router__pb2.HeartbeatRequest.FromString,
                    response_serializer=remote__router__pb2.HeartbeatResponse.SerializeToString,
            ),
            'get_version': grpc.unary_unary_rpc_method_handler(
                    servicer.get_version,
                    request_deserializer=remote__router__pb2.VersionRequest.FromString,
                    response_serializer=remote__router__pb2.VersionResponse.SerializeToString,
            ),
            'connect': grpc.unary_unary_rpc_method_handler(
                    servicer.connect,
                    request_deserializer=remote__router__pb2.ConnectRequest.FromString,
                    response_serializer=remote__router__pb2.ConnectResponse.SerializeToString,
            ),
            'reconnect': grpc.unary_unary_rpc_method_handler(
                    servicer.reconnect,
                    request_deserializer=remote__router__pb2.ReconnectRequest.FromString,
                    response_serializer=remote__router__pb2.ReconnectResponse.SerializeToString,
            ),
            'disconnect': grpc.unary_unary_rpc_method_handler(
                    servicer.disconnect,
                    request_deserializer=remote__router__pb2.DisconnectRequest.FromString,
                    response_serializer=remote__router__pb2.DisconnectResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'RemoteRouterService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class RemoteRouterService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def client_heartbeat(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/RemoteRouterService/client_heartbeat',
            remote__router__pb2.HeartbeatRequest.SerializeToString,
            remote__router__pb2.HeartbeatResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def get_version(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/RemoteRouterService/get_version',
            remote__router__pb2.VersionRequest.SerializeToString,
            remote__router__pb2.VersionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def connect(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/RemoteRouterService/connect',
            remote__router__pb2.ConnectRequest.SerializeToString,
            remote__router__pb2.ConnectResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def reconnect(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/RemoteRouterService/reconnect',
            remote__router__pb2.ReconnectRequest.SerializeToString,
            remote__router__pb2.ReconnectResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def disconnect(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/RemoteRouterService/disconnect',
            remote__router__pb2.DisconnectRequest.SerializeToString,
            remote__router__pb2.DisconnectResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



from typing import Optional

from aim import Run
from aim.ext.resource import DEFAULT_SYSTEM_TRACKING_INT

try:
    from lightgbm.callback import CallbackEnv
except ImportError:
    raise RuntimeError(
        'This contrib module requires Lightgbm to be installed. '
        'Please install it with command: \n pip install lightgbm'
    )


class AimCallback:
    """
    Records evaluation data into Aim.
    Similar to `lightgbm.record_evaluation` callback:
    https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.record_evaluation.html
    """

    def __init__(self, repo: Optional[str] = None,
                 experiment: Optional[str] = None,
                 system_tracking_interval: Optional[int]
                 = DEFAULT_SYSTEM_TRACKING_INT,
                 log_system_params: Optional[int] = True):
        self._repo_path = repo
        self._experiment = experiment
        self._system_tracking_interval = system_tracking_interval
        self._log_system_params = log_system_params
        self._run = None
        self._run_hash = None

        # callback parameters
        self.order = 25
        self.before_iteration = False

    @property
    def experiment(self) -> Run:
        if not self._run:
            self.setup()
        return self._run

    def setup(self):
        if self._run:
            return
        if self._run_hash:
            self._run = Run(
                self._run_hash,
                repo=self._repo_path,
                system_tracking_interval=self._system_tracking_interval,
            )
        else:
            self._run = Run(
                repo=self._repo_path,
                experiment=self._experiment,
                system_tracking_interval=self._system_tracking_interval,
                log_system_params=self._log_system_params
            )
            self._run_hash = self._run.hash

    def before_tracking(self, **kwargs):
        """ Runs before tracking data """
        pass

    def after_tracking(self, **kwargs):
        """ Runs after tracking data """
        pass

    def __call__(self, env: CallbackEnv):
        if env.iteration == env.begin_iteration:
            self.setup()

        self.before_tracking(env=env)

        for item in env.evaluation_result_list:
            if len(item) == 4:
                data_name, eval_name, result, _ = item
                self._run.track(result, name=eval_name, context={'data_name': data_name})
            else:
                data_name, eval_name = item[1].split()
                res_mean = item[2]
                res_stdv = item[4]
                self._run.track(res_mean, name=f'{eval_name}-mean', context={'data_name': data_name})
                self._run.track(res_stdv, name=f'{eval_name}-stdv', context={'data_name': data_name})

        self.after_tracking(env=env)

    def close(self):
        if self._run:
            self._run.close()
            del self._run
            self._run = None

    def __del__(self):
        self.close()



import time
import os

from tests.base import TestBase

from aim.sdk.run import Run
from aim.sdk.configs import AIM_ENABLE_TRACKING_THREAD


class TestRunFinalizedAt(TestBase):
    def _query_run_finalized_at(self, run_hash):
        run = self.repo.get_run(run_hash)
        return run.end_time

    def test_implicit_run_delete(self):
        run_hash = []

        def _func():
            run = Run(system_tracking_interval=None)
            run_hash.append(run.hash)
            self.assertIsNone(run.end_time)
            for i in range(10):
                run.track(i, name='seq')
            self.assertIsNone(run.end_time)

        _func()

        time.sleep(.1)  # wait for tracking queue and gc to collect the run object
        self.assertIsNotNone(self._query_run_finalized_at(run_hash[0]))

    def test_explicit_run_delete(self):
        run = Run(system_tracking_interval=None)
        run_hash = run.hash
        for i in range(10):
            run.track(i, name='seq')
        del run
        time.sleep(.1)
        self.assertIsNotNone(self._query_run_finalized_at(run_hash))

    def test_explicit_run_finalize(self):
        run = Run(system_tracking_interval=None)
        for i in range(10):
            run.track(i, name='seq')
        self.assertIsNone(run.end_time)
        run.finalize()
        self.assertIsNotNone(run.end_time)


class TestRunFinalizedAtWithTrackingQueue(TestRunFinalizedAt):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        Run.track_in_thread = True

    @classmethod
    def tearDownClass(cls) -> None:
        Run.track_in_thread = False
        super().tearDownClass()



from typing import Iterator, Tuple, Union

from typing import TYPE_CHECKING

from aim.storage import utils
from aim.storage.utils import interfaces

if TYPE_CHECKING:
    from aim.storage.treeview import TreeView
    from aim.storage.types import BLOB

ContainerKey = bytes
ContainerValue = Union[bytes, 'BLOB']


class Container:
    """Container is an interface of mutable key-value storage.

    We can think of :obj:`Container` as a set of `(key, value)` records. This
    definition enables us to easily define union-containers, store a Container
    physically in separate locations in chunks, optionally by partitioning data
    w.r.t. key contents (for example, group data chunks by key prefixes).

    The interface is similar to a flat dict-like object where both keys and the
    values are `bytes` objects. The design of interface makes it easy to manage
    groups of data to be stored with shared prefixes and implementations should
    be designed to provide optimized access to such constructs.
    """

    def close(self):
        """Close all the resources."""
        ...

    def finalize(self, index: 'Container'):
        """Finalize the Container.

        Perform operations of compactions, indexing, optimization, etc.
        """
        ...

    def preload(self):
        """Preload the Container.

        The interface of Container is designed in such a way that (almost) all
        the operations are supported to be done lazily.
        Sometimes there is need to preload the storage without performing an
        operation that will cause an actual read / write access.
        """
        ...

    def get(self, key: ContainerKey, default=None) -> ContainerValue:
        """Returns the value by the given `key` if it exists else `default`.

        The `default` is :obj:`None` by default.
        """
        ...

    def __getitem__(self, key: ContainerKey) -> ContainerValue:
        """Returns the value by the given `key`."""
        ...

    def set(self, key: ContainerKey, value: ContainerValue, store_batch=None):
        """Set a value for given key, optionally store in a batch.

        If `store_batch` is provided, instead of the `(key, value)` being added
        to the collection immediately, the operation is stored in a batch in
        order to be executed in a whole with other write operations. Depending
        on the :obj:`Container` implementation, this may feature transactions,
        atomic writes, etc.
        """
        ...

    def __setitem__(self, key: ContainerKey, value: ContainerValue):
        """Set a value for given key."""
        ...

    def delete(self, key: ContainerKey, store_batch=None):
        """Delete a key-value record by the given key,
        optionally store in a batch.

        If `store_batch` is provided, instead of the `(key, value)` being added
        to the collection immediately, the operation is stored in a batch in
        order to be executed in a whole with other write operations. Depending
        on the :obj:`Conainer` implementation, this may feature transactions,
        atomic writes, etc.
        """
        ...

    def __delitem__(self, key: ContainerKey):
        """Delete a key-value record by the given key."""
        ...

    def delete_range(self, begin: ContainerKey, end: ContainerKey, store_batch=None):
        """Delete all the records in the given `[begin, end)` key range."""
        ...

    def batch(self):
        """Creates a new batch object to store operations in before executing
        using :obj:`Container.commit`.

        The operations :obj:`Container.set`, :obj:`Container.delete`,
        :obj:`Container.delete_range` are supported.

        See more at :obj:`Container.commit`
        """
        ...

    def commit(self, batch):
        """Execute the accumulated write operations in the given `batch`.

        Depending on the :obj:`Container` implementation, this may feature
        transactions, atomic writes, etc.
        """
        ...

    def items(self, prefix: ContainerKey = b'') -> Iterator[Tuple[ContainerKey, ContainerValue]]:
        """Iterate over all the key-value records in the prefix key range.

        The iteration is always performed in lexicographic order w.r.t keys.
        If `prefix` is provided, iterate only over those records that have key
        starting with the `prefix`.

        For example, if `prefix == b'meta.'`, and the Container consists of:
        `{
            b'e.y': b'012',
            b'meta.x': b'123',
            b'meta.z': b'x',
            b'zzz': b'oOo'
        }`, the method will yield `(b'meta.x', b'123')` and `(b'meta.z', b'x')`

        Args:
            prefix (:obj:`bytes`): the prefix that defines the key range
        """
        ...

    def keys(self, prefix: ContainerKey = b'') -> Iterator[ContainerKey]:
        """Iterate over all the keys in the prefix range.

        The iteration is always performed in lexicographic order.
        If `prefix` is provided, iterate only over keys starting with
        the `prefix`.

        For example, if `prefix == b'meta.'`, and the Container consists of:
        `{
            b'e.y': b'012',
            b'meta.x': b'123',
            b'meta.z': b'x',
            b'zzz': b'oOo'
        }`, the method will yield `b'meta.x'` and `b'meta.z'`

        Args:
            prefix (:obj:`bytes`): the prefix that defines the key range
        """
        return utils.KeysIterator(self.items(prefix=prefix))

    def values(self, prefix: ContainerKey = b'') -> Iterator[ContainerValue]:
        """Iterate over all the values in the given prefix key range.

        The iteration is always performed in lexicographic order w.r.t keys.
        If `prefix` is provided, iterate only over those record values that have
        key starting with the `prefix`.

        For example, if `prefix == b'meta.'`, and the Container consists of:
        `{
            b'e.y': b'012',
            b'meta.x': b'123',
            b'meta.z': b'x',
            b'zzz': b'oOo'
        }`, the method will yield `b'123'` and `b'x'`

        Args:
            prefix (:obj:`bytes`): the prefix that defines the key range
        """
        return utils.ValuesIterator(self.items(prefix=prefix))

    def view(self, prefix: ContainerKey = b'') -> 'Container':
        """Return a view (even mutable ones) that enable access to the container
        but with modifications.

        Args:
            prefix (:obj:`bytes`): the prefix that defines the key range of the
                view-container. The resulting container will share an access to
                only records in the `prefix` key range, but with `prefix`-es
                stripped from them.

                For example, if the Container contents are:
                `{
                    b'e.y': b'012',
                    b'meta.x': b'123',
                    b'meta.z': b'x',
                    b'zzz': b'oOo'
                }`, then `container.view(prefix=b'meta.')` will behave (almost)
                exactly as an Container:
                `{
                    b'x': b'123',
                    b'z': b'x',
                }`
        """
        ...

    def tree(self) -> 'TreeView':
        """Return a :obj:`TreeView` which enables hierarchical view and access
        to the container records.

        This is achieved by prefixing groups and using `PATH_SENTINEL` as a
        separator for keys.

        For example, if the Container contents are:
            `{
                b'e.y': b'012',
                b'meta.x': b'123',
                b'meta.z': b'x',
                b'zzz': b'oOo'
            }`, and the path sentinel is `b'.'` then `tree = container.tree()`
            will behave as a (possibly deep) dict-like object:
            `tree[b'meta'][b'x'] == b'123'`
        """
        ...

    def walk(self, prefix: ContainerKey = b''):
        """A bi-directional generator to walk over the collection of records on
        any arbitrary order. The `prefix` sent to the generator (lets call it
        a `walker`) seeks for lower-bound key in the collection.

        In other words, if the Container contents are:
        `{
            b'e.y': b'012',
            b'meta.x': b'123',
            b'meta.z': b'x',
            b'zzz': b'oOo'
        }` and `walker = container.walk()` then:
        `walker.send(b'meta') == b'meta.x'`, `walker.send(b'e.y') == b'e.y'`
        """
        ...

    def next_key(self, key: ContainerKey = b'') -> ContainerKey:
        """Returns the key that comes (lexicographically) right after the
        provided `key`.
        """
        key, _ = self.next_item(key)
        return key

    def next_value(self, key: ContainerKey = b'') -> ContainerValue:
        """Returns the value for the key that comes (lexicographically) right
        after the provided `key`.
        """
        _, value = self.next_item(key)
        return value

    def next_key_value(self, key: ContainerKey = b'') -> Tuple[ContainerKey, ContainerValue]:
        """Returns `(key, value)` for the key that comes (lexicographically)
        right after the provided `key`.
        """
        return self.next_item(key)

    def next_item(self, key: ContainerKey = b'') -> Tuple[ContainerKey, ContainerValue]:
        """Returns `(key, value)` for the key that comes (lexicographically)
        right after the provided `key`.
        """
        ...

    def prev_key(self, key: ContainerKey = b'') -> ContainerKey:
        """Returns the key that comes (lexicographically) right before the
        provided `key`.
        """
        key, _ = self.prev_item()
        return key

    def prev_value(self, key: ContainerKey = b'') -> ContainerValue:
        """Returns the value for the key that comes (lexicographically) right
        before the provided `key`.
        """
        _, value = self.prev_item()
        return value

    def prev_key_value(self, key: ContainerKey = b'') -> Tuple[ContainerKey, ContainerValue]:
        """Returns `(key, value)` for the key that comes (lexicographically)
        right before the provided `key`.
        """
        return self.prev_item(key)

    def prev_item(self, key: ContainerKey = b'') -> Tuple[ContainerKey, ContainerValue]:
        """Returns `(key, value)` for the key that comes (lexicographically)
        right before the provided `key`.
        """
        ...


class ContainerItemsIterator(interfaces.Iterator):
    pass



import logging
from functools import wraps

logger = logging.getLogger(__name__)


def deprecated(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        logger.warning(msg=f'Usage of {func.__qualname__} is deprecated!')
        return func(*args, **kwargs)

    return wrapper



import numpy as np

from typing import Any, Iterator, List, Tuple, Union
from aim.storage.treeview import TreeView

from aim.storage.arrayview import ArrayView


class TreeArrayView(ArrayView):
    def __init__(
        self,
        tree: 'TreeView',
        dtype: Any = None
    ):
        self.tree = tree
        self.dtype = dtype

    def allocate(self):
        self.tree.make_array()
        return self

    def __iter__(self) -> Iterator[Any]:
        yield from self.values()

    def keys(self) -> Iterator[int]:
        yield from self.tree.keys()

    def indices(self) -> Iterator[int]:
        yield from self.keys()

    def values(self) -> Iterator[Any]:
        for k, v in self.tree.items():
            yield v

    def items(self) -> Iterator[Tuple[int, Any]]:
        yield from self.tree.items()

    def __len__(self) -> int:
        # TODO lazier
        try:
            last_idx = self.last_idx()
        except KeyError:
            return 0
        return last_idx + 1

    def __bool__(self) -> bool:
        return bool(len(self))

    def __getitem__(
        self,
        idx: Union[int, slice]
    ) -> Any:
        if isinstance(idx, slice):
            raise NotImplementedError
        return self.tree[idx]

    # TODO implement append

    def __setitem__(
        self,
        idx: int,
        val: Any
    ):
        assert isinstance(idx, int)
        self.tree[idx] = val

    def sparse_list(self) -> Tuple[List[int], List[Any]]:
        indices = []
        values = []
        for k, v in self.items():
            indices.append(k)
            values.append(v)

        return indices, values

    def indices_list(self) -> List[int]:
        return list(self.indices())

    def values_list(self) -> List[Any]:
        return list(self.values())

    def sparse_numpy(self) -> Tuple[np.ndarray, np.ndarray]:
        # TODO URGENT implement using cython
        indices_list, values_list = self.sparse_list()
        indices_array = np.array(indices_list, dtype=np.intp)
        values_array = np.array(values_list, dtype=self.dtype)
        return indices_array, values_array

    def indices_numpy(self) -> np.ndarray:
        # TODO URGENT implement using cython
        return np.array(self.indices_list(), dtype=np.intp)

    def values_numpy(self) -> np.ndarray:
        # TODO URGENT implement using cython
        return np.array(self.values_list(), dtype=self.dtype)

    def tolist(self) -> List[Any]:
        arr = self.tree[...]
        assert isinstance(arr, list)
        return arr

    def first(self) -> Tuple[int, Any]:
        idx = self.first_idx()
        return idx, self[idx]

    def first_idx(self) -> int:
        return self.tree.first_key()

    def first_value(self) -> Any:
        return self[self.first_idx()]

    def last(self) -> Tuple[int, Any]:
        idx = self.last_idx()
        return idx, self[idx]

    def last_idx(self) -> int:
        return self.tree.last_key()

    def last_value(self) -> Any:
        return self[self.last_idx()]



from logging import getLogger
from typing import Optional

from aim.ext.resource.configs import DEFAULT_SYSTEM_TRACKING_INT
from aim.sdk.num_utils import is_number
from aim.sdk.run import Run

try:
    from transformers.trainer_callback import TrainerCallback
except ImportError:
    raise RuntimeError(
        'This contrib module requires Transformers to be installed. '
        'Please install it with command: \n pip install transformers'
    )

logger = getLogger(__name__)


class AimCallback(TrainerCallback):
    def __init__(self,
                 repo: Optional[str] = None,
                 experiment: Optional[str] = None,
                 system_tracking_interval: Optional[int]
                 = DEFAULT_SYSTEM_TRACKING_INT,
                 log_system_params: bool = True,
                 ):
        self._repo_path = repo
        self._experiment_name = experiment
        self._system_tracking_interval = system_tracking_interval
        self._log_system_params = log_system_params
        self._run = None
        self._run_hash = None
        self._log_value_warned = False

    @property
    def experiment(self):
        if not self._run:
            self.setup()
        return self._run

    def setup(self, args=None, state=None, model=None):
        if state and not state.is_world_process_zero:
            return

        if not self._run:
            if self._run_hash:
                self._run = Run(
                    self._run_hash,
                    repo=self._repo_path,
                    system_tracking_interval=self._system_tracking_interval,
                )
            else:
                self._run = Run(
                    repo=self._repo_path,
                    experiment=self._experiment_name,
                    system_tracking_interval=self._system_tracking_interval,
                    log_system_params=self._log_system_params,
                )
                self._run_hash = self._run.hash

        if args:
            combined_dict = {**args.to_sanitized_dict()}
            for key, value in combined_dict.items():
                self._run.set(('hparams', key), value, strict=False)

        # Store model configs as well
        # if hasattr(model, 'config') and model.config is not None:
        #     model_config = model.config.to_dict()
        #     self._run['model'] = model_config

    def on_train_begin(self, args, state, control,
                       model=None, **kwargs):
        if not state.is_world_process_zero:
            return
        if not self._run:
            self.setup(args, state, model)

    def on_train_end(self, args, state, control, **kwargs):
        if not state.is_world_process_zero:
            return
        self.close()

    def on_log(self, args, state, control,
               model=None, logs=None, **kwargs):
        if not state.is_world_process_zero:
            return

        if not self._run:
            self.setup(args, state, model)

        for log_name, log_value in logs.items():
            context = {}
            prefix_set = {'train_', 'eval_', 'test_'}
            for prefix in prefix_set:
                if log_name.startswith(prefix):
                    log_name = log_name[len(prefix):]
                    context = {'subset': prefix[:-1]}
                    break
            if not is_number(log_value):
                if not self._log_value_warned:
                    self._log_value_warned = True
                    logger.warning(
                        f'Trainer is attempting to log a value of '
                        f'"{log_value}" of type {type(log_value)} for key "{log_name}"'
                        f' as a metric which is not a supported value type.'
                    )
                continue

            self._run.track(log_value,
                            name=log_name, context=context,
                            step=state.global_step, epoch=state.epoch)

    def close(self):
        if self._run:
            self._run.close()
            del self._run
            self._run = None

    def __del__(self):
        self.close()



import os
import logging
from aim.web.configs import AIM_LOG_LEVEL_KEY


def set_log_level(log_level):
    numeric_level = getattr(logging, log_level.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError('Invalid log level: %s' % log_level)
    os.environ[AIM_LOG_LEVEL_KEY] = str(numeric_level)
    logging.basicConfig(level=numeric_level)



import unittest
import os
from unittest import mock
from unittest.mock import MagicMock

import aim.sdk.utils as utils


class TestUtils(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_should_search_aim_repo_not_found_back_slash(self):
        # arrange
        path = '/'

        # act
        path, found = utils.search_aim_repo(path)

        # assert
        self.assertFalse(found)
        self.assertIsNone(path)

    @mock.patch('os.path.exists')
    def test_should_search_aim_repo_not_found_dot_back_slash(self, mock_os_path_exists: mock.MagicMock):
        # arrange
        path = './'
        built_dir = os.path.dirname(__file__)
        build_dir_split = built_dir.split(os.sep)
        mock_os_path_exists.side_effect = [False] * len(build_dir_split)

        # act
        path, found = utils.search_aim_repo(path)

        # assert
        self.assertFalse(found)
        self.assertIsNone(path)

    def test_should_generate_run_hash_default_len(self):
        # arrange
        expected_hash_len = 24

        # act
        actual_hash = utils.generate_run_hash()

        # assert
        self.assertIsInstance(actual_hash, str)
        self.assertEqual(expected_hash_len, len(actual_hash))

    def test_should_generate_run_hash_twelve_char_long(self):
        # arrange
        hash_length = 12
        expected_hash_len = 12

        # act
        actual_hash = utils.generate_run_hash(hash_length)

        # assert
        self.assertIsInstance(actual_hash, str)
        self.assertEqual(expected_hash_len, len(actual_hash))

    @mock.patch('uuid.uuid4')
    def test_should_generate_run_hash_six_char_long(self, mock_uuid: MagicMock):
        # arrange
        expected_uuid_hex = '16710b81-ccab-4409-bd79-50a770b565a6'
        mock_uuid.return_value.hex = expected_uuid_hex
        hash_length = 6
        expected_hash_len = 6

        # act
        actual_hash = utils.generate_run_hash(hash_length)

        # assert
        self.assertIsInstance(actual_hash, str)
        self.assertEqual(expected_hash_len, len(actual_hash))
        self.assertEqual(expected_uuid_hex[:expected_hash_len], actual_hash)
        mock_uuid.assert_called_once()

    def test_should_get_obj_type_name_str(self):
        # arrange
        obj = 'hello aim!'
        expected_type = 'str'

        # act
        actual_type = utils.get_object_typename(obj)

        # assert
        self.assertEqual(expected_type, actual_type)

    def test_should_get_clean_repo_path(self):
        # arrange
        path = './'
        expected_repo_path = os.getcwd()

        # act
        actual_repo_path = utils.clean_repo_path(path)

        # assert
        self.assertEqual(expected_repo_path, actual_repo_path)

    def test_should_get_clean_repo_path_empty_str(self):
        # arrange
        path = ''
        expected_repo_path = ''

        # act
        actual_repo_path = utils.clean_repo_path(path)

        # assert
        self.assertEqual(expected_repo_path, actual_repo_path)

    def test_should_get_obj_type_name_list_unknown(self):
        # arrange
        obj = [None]
        expected_type = 'list(unknown)'

        # act
        actual_type = utils.get_object_typename(obj)

        # assert
        self.assertEqual(expected_type, actual_type)

    def test_should_get_obj_type_name_bool(self):
        # arrange
        obj = True
        expected_type = 'int'

        # act
        actual_type = utils.get_object_typename(obj)

        # assert
        self.assertEqual(expected_type, actual_type)

    def test_should_check_types_compatibility_int(self):
        # arrange
        data_type = 'int'
        base_data_type = 'int'

        # act
        actual_compatibility = utils.check_types_compatibility(data_type, base_data_type, None)

        # assert
        self.assertTrue(actual_compatibility)

    def test_should_check_types_compatibility_base_list(self):
        # arrange
        data_type = 'list(str)'
        base_data_type = 'list'

        # act
        actual_compatibility = utils.check_types_compatibility(data_type, base_data_type, None)

        # assert
        self.assertTrue(actual_compatibility)

    def test_should_check_types_compatibility_base_list_str(self):
        # arrange
        data_type = 'list'
        base_data_type = 'list(str)'

        # act
        actual_compatibility = utils.check_types_compatibility(data_type, base_data_type, None)

        # assert
        self.assertTrue(actual_compatibility)



import os
import click

from aim.cli.utils import set_log_level
from aim.cli.up.utils import build_db_upgrade_command, build_uvicorn_command, get_free_port_num
from aim.web.configs import (
    AIM_ENV_MODE_KEY,
    AIM_TF_LOGS_PATH_KEY,
    AIM_UI_BASE_PATH,
    AIM_UI_DEFAULT_HOST,
    AIM_UI_DEFAULT_PORT,
    AIM_UI_MOUNTED_REPO_PATH,
    AIM_UI_TELEMETRY_KEY,
    AIM_PROXY_URL,
    AIM_PROFILER_KEY
)
from aim.sdk.repo import Repo, RepoStatus
from aim.sdk.utils import clean_repo_path
from aim.web.utils import exec_cmd
from aim.web.utils import ShellCommandException


@click.command()
@click.option('-h', '--host', default=AIM_UI_DEFAULT_HOST, type=str)
@click.option('-p', '--port', default=AIM_UI_DEFAULT_PORT, type=int)
@click.option('-w', '--workers', default=1, type=int)
@click.option('--uds', required=False, type=click.Path(exists=False,
                                                       file_okay=True,
                                                       dir_okay=False,
                                                       readable=True))
@click.option('--repo', required=False, type=click.Path(exists=True,
                                                        file_okay=False,
                                                        dir_okay=True,
                                                        writable=True))
@click.option('--tf_logs', type=click.Path(exists=True, readable=True))
@click.option('--dev', is_flag=True, default=False)
@click.option('--ssl-keyfile', required=False, type=click.Path(exists=True,
                                                               file_okay=True,
                                                               dir_okay=False,
                                                               readable=True))
@click.option('--ssl-certfile', required=False, type=click.Path(exists=True,
                                                                file_okay=True,
                                                                dir_okay=False,
                                                                readable=True))
@click.option('--base-path', required=False, default='', type=str)
@click.option('--force-init', is_flag=True, default=False)
@click.option('--profiler', is_flag=True, default=False)
@click.option('--log-level', required=False, default='', type=str)
def up(dev, host, port, workers, uds,
       repo, tf_logs,
       ssl_keyfile, ssl_certfile,
       base_path, force_init,
       profiler, log_level):
    if dev:
        os.environ[AIM_ENV_MODE_KEY] = 'dev'
        log_level = log_level or 'debug'
    else:
        os.environ[AIM_ENV_MODE_KEY] = 'prod'

    if log_level:
        set_log_level(log_level)

    if base_path:
        # process `base_path` as ui requires leading slash
        if base_path.endswith('/'):
            base_path = base_path[:-1]
        if base_path and not base_path.startswith('/'):
            base_path = f'/{base_path}'
        os.environ[AIM_UI_BASE_PATH] = base_path

    repo_path = clean_repo_path(repo) or Repo.default_repo_path()
    repo_status = Repo.check_repo_status(repo_path)
    if repo_status == RepoStatus.MISSING:
        init_repo = None
        if not force_init:
            init_repo = click.confirm(f'\'{repo_path}\' is not a valid Aim repository. Do you want to initialize it?')
        else:
            init_repo = True
        if not init_repo:
            click.echo('To initialize repo please run the following command:')
            click.secho('aim init', fg='yellow')
            return
        repo_inst = Repo.from_path(repo_path, init=True)
    elif repo_status == RepoStatus.UPDATE_REQUIRED:
        upgrade_repo = click.confirm(f'\'{repo_path}\' requires upgrade. Do you want to run upgrade automatically?')
        if upgrade_repo:
            from aim.cli.upgrade.utils import convert_2to3
            repo_inst = convert_2to3(repo_path, drop_existing=False, skip_failed_runs=False, skip_checks=False)
        else:
            click.echo('To upgrade repo please run the following command:')
            click.secho(f'aim upgrade --repo {repo_path} 2to3', fg='yellow')
            return
    else:
        repo_inst = Repo.from_path(repo_path)

    os.environ[AIM_UI_MOUNTED_REPO_PATH] = repo_inst.path

    if tf_logs:
        os.environ[AIM_TF_LOGS_PATH_KEY] = tf_logs

    try:
        db_cmd = build_db_upgrade_command()
        exec_cmd(db_cmd, stream_output=True)
    except ShellCommandException:
        click.echo('Failed to initialize Aim DB. '
                   'Please see the logs above for details.')
        return

    if port == 0:
        try:
            port = get_free_port_num()
        except Exception:
            pass

    if dev or (os.getenv(AIM_UI_TELEMETRY_KEY) is not None and os.getenv(AIM_UI_TELEMETRY_KEY) == '0'):
        os.environ[AIM_UI_TELEMETRY_KEY] = '0'
    else:
        os.environ[AIM_UI_TELEMETRY_KEY] = '1'
        alert_msg = 'Aim UI collects anonymous usage analytics.'
        opt_out_msg = 'Read how to opt-out here: '
        opt_out_url = 'https://aimstack.readthedocs.io/en/latest/community/telemetry.html'
        line_width = max(len(opt_out_msg), len(alert_msg), len(opt_out_url)) + 8
        click.echo('-' * line_width)
        click.echo('{}{}{}'.format(' ' * ((line_width - len(alert_msg)) // 2),
                                   alert_msg,
                                   ' ' * ((line_width - len(alert_msg)) // 2)))
        click.echo('{}{}{}'.format(' ' * ((line_width - len(opt_out_msg)) // 2),
                                   opt_out_msg,
                                   ' ' * ((line_width - len(opt_out_msg)) // 2)))
        click.echo('{}{}{}'.format(' ' * ((line_width - len(opt_out_url)) // 2),
                                   opt_out_url,
                                   ' ' * ((line_width - len(opt_out_url)) // 2)))
        click.echo('-' * line_width)

    click.echo(click.style('Running Aim UI on repo `{}`'.format(repo_inst), fg='yellow'))

    if uds:
        click.echo('Aim UI running on {}'.format(uds))
    else:
        scheme = 'https' if ssl_keyfile or ssl_certfile else 'http'
        click.echo('Open {}://{}:{}{}'.format(scheme, host, port, base_path), err=True)

    proxy_url = os.environ.get(AIM_PROXY_URL)
    if proxy_url:
        click.echo(f'Proxy {proxy_url}{base_path}/')

    click.echo('Press Ctrl+C to exit')

    if profiler:
        os.environ[AIM_PROFILER_KEY] = '1'

    try:
        server_cmd = build_uvicorn_command(host, port, workers, uds, ssl_keyfile, ssl_certfile, log_level)
        exec_cmd(server_cmd, stream_output=True)
    except ShellCommandException:
        click.echo('Failed to run Aim UI. Please see the logs above for details.')
        return



from typing import Any, Callable, Dict, List, Optional, Union

from aim import Run
from aim.ext.resource.configs import DEFAULT_SYSTEM_TRACKING_INT

try:
    from torch.optim import Optimizer
except ImportError:
    raise RuntimeError(
        'This contrib module requires PyTorch to be installed. '
        'Please install it with command: \n pip install torch'
    )
try:
    from ignite.contrib.handlers.base_logger import BaseLogger, BaseOptimizerParamsHandler, BaseOutputHandler
    from ignite.engine import Engine, Events
except ImportError:
    raise RuntimeError(
        'This contrib module requires PyTorch Ignite to be installed. '
        'Please install it with command: \n pip install pytorch-ignite'
    )


class AimLogger(BaseLogger):
    """
        `Aim` tracking client handler to log parameters and metrics during the training
        and validation.

        Args:
            repo: Aim repository path
            experiment: Aim run's experiment name
            train_metric_prefix: prefix to identify training metrics
            val_metric_prefix: prefix to identify validation metrics
            test_metric_prefix: prefix to identify test metrics
            system_tracking_interval: sets the tracking interval in seconds
                for system usage metrics (CPU, Memory, etc.)
            log_system_params: flag to enable system params logging, True by default
    """
    def __init__(self,
                 repo: Optional[str] = None,
                 experiment: Optional[str] = None,
                 train_metric_prefix: Optional[str] = 'train_',
                 val_metric_prefix: Optional[str] = 'val_',
                 test_metric_prefix: Optional[str] = 'test_',
                 system_tracking_interval: Optional[int]
                 = DEFAULT_SYSTEM_TRACKING_INT,
                 log_system_params: bool = True,
                 ):
        super().__init__()

        self._experiment_name = experiment
        self._repo_path = repo

        self._train_metric_prefix = train_metric_prefix
        self._val_metric_prefix = val_metric_prefix
        self._test_metric_prefix = test_metric_prefix
        self._system_tracking_interval = system_tracking_interval
        self._log_system_params = log_system_params

        self._run = None
        self._run_hash = None

    @property
    def experiment(self) -> Run:
        if self._run is None:
            if self._run_hash:
                self._run = Run(
                    self._run_hash,
                    repo=self._repo_path,
                    system_tracking_interval=self._system_tracking_interval,
                )
            else:
                self._run = Run(
                    repo=self._repo_path,
                    experiment=self._experiment_name,
                    system_tracking_interval=self._system_tracking_interval,
                    log_system_params=self._log_system_params,
                )
                self._run_hash = self._run.hash
        return self._run

    def log_params(self, params: dict):
        # Handle OmegaConf object
        try:
            from omegaconf import OmegaConf, Container
        except ModuleNotFoundError:
            pass
        else:
            # Convert to primitives
            if isinstance(params, Container):
                params = OmegaConf.to_container(params, resolve=True)

        for key, value in params.items():
            self.experiment.set(('hparams', key), value, strict=False)

    def log_metrics(self, metrics: Dict[str, float],
                    step: Optional[int] = None):
        for k, v in metrics.items():
            name = k
            context = {}
            if self._train_metric_prefix \
                    and name.startswith(self._train_metric_prefix):
                name = name[len(self._train_metric_prefix):]
                context['subset'] = 'train'
            elif self._test_metric_prefix \
                    and name.startswith(self._test_metric_prefix):
                name = name[len(self._test_metric_prefix):]
                context['subset'] = 'test'
            elif self._val_metric_prefix \
                    and name.startswith(self._val_metric_prefix):
                name = name[len(self._val_metric_prefix):]
                context['subset'] = 'val'
            self.experiment.track(v, step=step, name=name, context=context)

    @property
    def save_dir(self) -> str:
        return self.experiment.repo.path

    @property
    def name(self) -> str:
        return self._experiment_name

    @property
    def version(self) -> str:
        return self.experiment.hash

    def close(self) -> None:
        if self._run:
            self._run.close()
            del self._run
            self._run = None

    def __del__(self):
        self.close()

    def _create_output_handler(self, *args: Any, **kwargs: Any) -> "OutputHandler":
        return OutputHandler(*args, **kwargs)

    def _create_opt_params_handler(self, *args: Any, **kwargs: Any) -> "OptimizerParamsHandler":
        return OptimizerParamsHandler(*args, **kwargs)


class OutputHandler(BaseOutputHandler):
    """Helper handler to log engine's output and/or metrics.

        Args:
            tag: common title for all produced plots. For example, 'training'
            metric_names: list of metric names to plot or a string "all" to plot all available
                metrics.
            output_transform: output transform function to prepare `engine.state.output` as a number.
                For example, `output_transform = lambda output: output`
                This function can also return a dictionary, e.g `{'loss': loss1, 'another_loss': loss2}` to label
                the plot with corresponding keys.
            global_step_transform: global step transform function to output a desired global step.
                Input of the function is `(engine, event_name)`. Output of function should be an integer.
                Default is None, global_step based on attached engine. If provided,
                uses function output as global_step. To setup global step from another engine, please use
                :meth:`~ignite.contrib.handlers.mlflow_logger.global_step_from_engine`.
            state_attributes: list of attributes of the ``trainer.state`` to plot.
    """
    def __init__(
        self,
        tag: str,
        metric_names: Optional[Union[str, List[str]]] = None,
        output_transform: Optional[Callable] = None,
        global_step_transform: Optional[Callable] = None,
        state_attributes: Optional[List[str]] = None,
    ) -> None:
        super(OutputHandler, self).__init__(
            tag, metric_names, output_transform, global_step_transform, state_attributes
        )

    def __call__(self, engine: Engine, logger: AimLogger, event_name: Union[str, Events]) -> None:

        if not isinstance(logger, AimLogger):
            raise TypeError('Handler \'OutputHandler\' works only with AimLogger')

        rendered_metrics = self._setup_output_metrics_state_attrs(engine)

        global_step = self.global_step_transform(engine, event_name)  # type: ignore[misc]

        if not isinstance(global_step, int):
            raise TypeError(
                f'global_step must be int, got {type(global_step)}.'
                ' Please check the output of global_step_transform.'
            )

        metrics = {}
        for keys, value in rendered_metrics.items():
            key = '_'.join(keys)
            metrics[key] = value

        logger.log_metrics(metrics, step=global_step)


class OptimizerParamsHandler(BaseOptimizerParamsHandler):
    """Helper handler to log optimizer parameters

        Args:
            optimizer: torch optimizer or any object with attribute ``param_groups``
                as a sequence.
            param_name: parameter name
            tag: common title for all produced plots. For example, 'generator'
    """
    def __init__(self, optimizer: Optimizer, param_name: str = 'lr', tag: Optional[str] = None):
        super(OptimizerParamsHandler, self).__init__(optimizer, param_name, tag)

    def __call__(self, engine: Engine, logger: AimLogger, event_name: Union[str, Events]) -> None:
        if not isinstance(logger, AimLogger):
            raise TypeError('Handler OptimizerParamsHandler works only with AimLogger')

        global_step = engine.state.get_event_attrib_value(event_name)
        tag_prefix = f'{self.tag}_' if self.tag else ''
        params = {
            f'{tag_prefix}{self.param_name}_group_{i}': float(param_group[self.param_name])
            for i, param_group in enumerate(self.optimizer.param_groups)
        }

        logger.log_metrics(params, step=global_step)



import logging
import time
from typing import List, Optional

from aim.ext.notifier.base_notifier import BaseNotifier

logger = logging.getLogger(__name__)


class NotificationSendError(RuntimeError):
    def __init__(self, e: Exception):
        super().__init__(e)


class Notifier(BaseNotifier):
    MAX_RETRIES = 5
    RETRY_DELAY = 1

    def __init__(self):
        self._notifiers: List[BaseNotifier] = []

    def add(self, sub: BaseNotifier):
        self._notifiers.append(sub)

    def notify(self, message: Optional[str] = None, **kwargs):
        for sub in self._notifiers:
            attempt = 0
            while attempt < self.MAX_RETRIES:
                try:
                    sub.notify(message, **kwargs)
                    break
                except Exception as e:
                    attempt += 1
                    if attempt == self.MAX_RETRIES:
                        logger.error(f'Notifier {sub} failed to send message "{message}". '
                                     f'No retries left.')
                        raise NotificationSendError(e)
                    else:
                        logger.error(f'Notifier {sub} failed to send message "{message}". '
                                     f'Retry attempts left {self.MAX_RETRIES - attempt} '
                                     f'Next retry in {self.RETRY_DELAY} seconds.')
                        time.sleep(self.RETRY_DELAY)



import click
import os
from tqdm import tqdm

from aim.cli.runs.utils import list_repo_runs, match_runs
from aim.cli.upgrade.utils import convert_2to3

from aim.sdk.maintenance_run import MaintenanceRun as Run
from aim.sdk.utils import backup_run, restore_run_backup, clean_repo_path
from aim.sdk.repo import Repo
from aim.sdk.index_manager import RepoIndexManager


@click.group()
@click.option('--repo', required=False,
              default=os.getcwd(),
              type=click.Path(
                  exists=True,
                  file_okay=False,
                  dir_okay=True,
                  writable=True
              ))
@click.pass_context
def storage(ctx, repo):
    """Manage aim repository data & format updates."""
    ctx.ensure_object(dict)
    ctx.obj['repo'] = repo


@storage.group()
@click.pass_context
def upgrade(ctx):
    """Update Runs data for given run hashes to use new format."""
    pass


@upgrade.command(name='2to3')
@click.option('--drop-existing', required=False, is_flag=True, default=False)
@click.option('--skip-failed-runs', required=False, is_flag=True, default=False)
@click.option('--skip-checks', required=False, is_flag=True, default=False)
@click.pass_context
def v2to3(ctx, drop_existing, skip_failed_runs, skip_checks):
    repo_path = ctx.obj['repo']
    convert_2to3(repo_path, drop_existing, skip_failed_runs, skip_checks)


@upgrade.command(name='3.11+')
@click.argument('hashes', nargs=-1, type=str)
@click.pass_context
def to_3_11(ctx, hashes):
    """Optimize Runs Metrics data for read access."""
    if len(hashes) == 0:
        click.echo('Please specify at least one Run to update.')
        exit(1)
    repo_path = ctx.obj['repo']
    repo = Repo.from_path(repo_path)

    matched_hashes = match_runs(repo_path, hashes)
    remaining_runs = []
    confirmed = click.confirm(f'This command will optimize the metrics data for {len(matched_hashes)} runs from aim '
                              f'repo located at \'{repo_path}\'. This process might take a while. '
                              f'Do you want to proceed?')
    if not confirmed:
        return

    index_manager = RepoIndexManager.get_index_manager(repo)
    for run_hash in tqdm(matched_hashes):
        try:
            run = Run(run_hash, repo=repo)
            if run.check_metrics_version():
                backup_run(run)
                run.update_metrics()
                index_manager.index(run_hash)
            else:
                click.echo(f'Run {run.hash} is already up to date. Skipping')
        except Exception:
            remaining_runs.append(run_hash)

    if not remaining_runs:
        click.echo('Finished optimizing metric data.')
    else:
        click.echo('Finished optimizing metric data. The following runs were skipped:')
        click.secho(' '.join(remaining_runs), fg='yellow')
    click.echo('In case of any issues the following command can be used to restore data:')
    click.secho(f'aim storage --repo {repo.root_path} restore \'*\'', fg='yellow')


@storage.command(name='restore')
@click.argument('hashes', nargs=-1, type=str)
@click.pass_context
def restore_runs(ctx, hashes):
    """Rollback Runs data for given run hashes to the previous metric format. """
    if len(hashes) == 0:
        click.echo('Please specify at least one Run to delete.')
        exit(1)
    repo_path = ctx.obj['repo']
    repo = Repo.from_path(repo_path)

    matched_hashes = match_runs(repo_path, hashes, lookup_dir='bcp')
    confirmed = click.confirm(f'This command will restore {len(matched_hashes)} runs from aim repo '
                              f'located at \'{repo_path}\'. Do you want to proceed?')
    if not confirmed:
        return

    remaining_runs = []
    index_manager = RepoIndexManager.get_index_manager(repo)
    for run_hash in tqdm(matched_hashes):
        try:
            restore_run_backup(repo, run_hash)
            index_manager.index(run_hash)
        except Exception as e:
            click.echo(f'Error while trying to restore run \'{run_hash}\'. {str(e)}.', err=True)
            remaining_runs.append(run_hash)

    if not remaining_runs:
        click.echo(f'Successfully restored {len(matched_hashes)} runs.')
    else:
        click.echo('Something went wrong while restoring runs. Remaining runs are:', err=True)
        click.secho('\t'.join(remaining_runs), fg='yellow')


@storage.command(name='prune')
@click.pass_context
def prune(ctx):
    """Remove dangling/orphan params/sequences with no referring runs."""

    from collections.abc import MutableMapping

    def flatten(d, parent_path=None):
        if parent_path and not isinstance(parent_path, tuple):
            parent_path = (parent_path, )

        all_paths = set()
        for k, v in d.items():
            if k == '__example_type__':
                continue

            new_path = parent_path + (k,) if parent_path else (k, )
            all_paths.add(new_path)
            if isinstance(v, MutableMapping):
                all_paths.update(flatten(v, new_path))

        return all_paths

    repo_path = ctx.obj['repo']
    repo = Repo.from_path(repo_path)

    subtrees_to_lookup = ('attrs', 'traces_types', 'contexts', 'traces')
    repo_meta_tree = repo._get_meta_tree()

    # set of all repo paths that can be left dangling after run deletion
    repo_paths = set()
    for key in subtrees_to_lookup:
        try:
            repo_paths.update(flatten(repo_meta_tree.collect(key, strict=False), parent_path=(key,)))
        except KeyError:
            pass

    run_hashes = list_repo_runs(clean_repo_path(repo.path))
    for run_hash in tqdm(run_hashes):
        # construct unique paths set for each run
        run_paths = set()
        run_meta_tree = repo.request_tree('meta', run_hash, from_union=False, read_only=True).subtree('meta')
        for key in subtrees_to_lookup:
            try:
                run_paths.update(flatten(run_meta_tree.collect(key, strict=False), parent_path=(key,)))
            except KeyError:
                pass

        # update repo_paths keeping the elements that were not found in run_paths
        repo_paths.difference_update(run_paths)

        # if no paths are left in repo_paths set, means that we have no orphan paths
        if not repo_paths:
            break

    # everything left in the `repo_paths` set is subject to be deleted
    if not repo_paths:
        click.echo('No orphan params were found')
        return

    # acquire index container to delete orphan paths
    index_tree = repo._get_index_tree('meta', timeout=5).subtree('meta')

    # start deleting with the deepest paths first to bypass the cases when parent path is deleted before the child
    for path in sorted(repo_paths, key=len, reverse=True):
        del index_tree[path]


@storage.command('reindex')
@click.option('--finalize-only', required=False, is_flag=True, default=False)
@click.pass_context
def reindex(ctx, finalize_only):
    """ Process runs left in 'in progress' state. """
    from aim.utils.deprecation import deprecation_warning

    deprecation_warning(remove_version='3.16', msg='`aim storage reindex` is deprecated! '
                                                   'Use `aim runs close` command instead.')
    return



import time
from abc import abstractmethod

import atexit
import weakref
import logging
import threading

from typing import Generic, Tuple, TypeVar, Dict

T = TypeVar('T')

logger = logging.getLogger(__name__)


class RobustExec(threading.Thread):
    """
    A thread that executes a function and catches all soft interrupts.
    Users very often Ctrl-C to stop the program multiple times which leaves
    program no chance to clean up.
    """
    def __init__(self, *args, stop_signal, **kwargs):
        super().__init__(*args, **kwargs)
        self.stop_signal = stop_signal

    def join(self):
        while True:
            try:
                # Pressing Ctrl-C will raise KeyboardInterrupt and stop only
                # the `Event.wait()` call. The target function of `RobustExec`
                # will continue to run, until stop_signal is set.
                return self.stop_signal.wait()
            except KeyboardInterrupt:
                logger.warning('Received Ctrl-C. Closing gracefully.')


class AutoClean(Generic[T]):
    PRIORITY = 10
    _registered_with_atexit = False
    _finalizers: Dict[
        T, Tuple[int, weakref.finalize]
    ] = weakref.WeakKeyDictionary()

    stop_signal = threading.Event()

    def __init__(self, instance: T) -> None:
        """
        Initialize object resources that need to be cleaned up automatically.
        """
        self.finalizer = weakref.finalize(instance, self._close)
        # We do not finalizer to automatically close the object itself.
        # Instead, we will collect all finalizers and call them at exit.
        self.finalizer.atexit = False
        self._finalizers[instance] = (self.PRIORITY, self.finalizer)

        if not AutoClean._registered_with_atexit:
            atexit.register(AutoClean.cleanup)
            AutoClean._registered_with_atexit = True

    def close(self) -> None:
        """Manually close the object."""
        self.finalizer()

    @abstractmethod
    def _close(self):
        """
        This method is called when the instance is garbage collected.
        Meant to be overridden by subclasses.
        """
        ...

    @staticmethod
    def _cleanup():
        """
        Actually cleanup all the remaining resources in order of specified
        priorities. Resources with bigger priority numbers are cleaned up first (have higher priority)
        while the resources with equal priorities are cleaned up in reverse order
        of their registration (hence why reversed is used).
        This is called automatically by atexit. It is not meant tobe called manually.
        """

        finalizers = sorted(AutoClean._finalizers.items(), key=lambda x: x[1][0])
        logger.debug(f'Cleaning up...  Found {len(finalizers)} finalizers')
        logger.debug('Cleaning up...  Iterating over instances in order')
        for key, (priority, finalizer) in reversed(finalizers):
            logger.debug(f'Cleaning up...  with priority={priority} instance {key}')
            finalizer()
        AutoClean.stop_signal.set()

    @staticmethod
    def cleanup():
        """
        Cleanup all the remaining objects. This is called automatically
        when the program exits.
        It also blocks until all the cleanup functions have finished to ensure
        data consistency.
        """
        logger.debug('Cleaning up...  Blocking KeyboardInterrupts')
        example = RobustExec(stop_signal=AutoClean.stop_signal, target=AutoClean._cleanup)
        example.start()
        example.join()
        logger.debug('Cleaning up...  Done')



from sys import stdout
from typing import Optional

from aim import Run
from aim.ext.resource import DEFAULT_SYSTEM_TRACKING_INT


class AimLogger:

    def __init__(self, loss_function: Optional[str] = "Loss",
                 repo: Optional[str] = None,
                 experiment: Optional[str] = None,
                 system_tracking_interval: Optional[int]
                 = DEFAULT_SYSTEM_TRACKING_INT,
                 log_system_params: Optional[bool] = True,
                 log_cout=stdout):
        super().__init__()
        self._loss_function = loss_function
        self._repo_path = repo
        self._experiment = experiment
        self._system_tracking_interval = system_tracking_interval
        self._log_system_params = log_system_params
        self._run = None
        self._run_hash = None
        self._log_cout = log_cout

        if log_cout is not None:
            assert hasattr(log_cout, "write")

    @property
    def experiment(self) -> Run:
        if not self._run:
            self.setup()
        return self._run

    def setup(self):
        if self._run:
            return
        if self._run_hash:
            self._run = Run(
                self._run_hash,
                repo=self._repo_path,
                system_tracking_interval=self._system_tracking_interval,
            )
        else:
            self._run = Run(
                repo=self._repo_path,
                experiment=self._experiment,
                system_tracking_interval=self._system_tracking_interval,
                log_system_params=self._log_system_params
            )
            self._run_hash = self._run.hash

    def _to_number(self, val):
        try:
            return float(val)
        except ValueError:
            return val

    def write(self, log):
        run = self.experiment

        _log = log
        log = log.strip().split()
        if log:
            if len(log) == 3 and log[1] == '=':
                run[log[0]] = self._to_number(log[2])
                return

            value_learn = None
            value_iter = None
            value_test = None
            value_best = None

            if log[1] == 'learn:':
                value_iter = int(log[0][:-1])
                value_learn = self._to_number(log[2])
                if log[3] == 'test:':
                    value_test = self._to_number(log[4])
                    if log[5] == 'best:':
                        value_best = self._to_number(log[6])
            if any((value_learn, value_test, value_best)):
                if value_learn:
                    run.track(value_learn, name=self._loss_function, step=value_iter, context={'log': 'learn'})
                if value_test:
                    run.track(value_test, name=self._loss_function, step=value_iter, context={'log': 'test'})
                if value_best:
                    run.track(value_best, name=self._loss_function, step=value_iter, context={'log': 'best'})
            else:
                # Unhandled or junky log
                pass

        if self._log_cout:
            self._log_cout.write(_log)



# distutils: language = c++
# cython: wraparound = False
# cython: boundscheck = False
# cython: cdivision=True
# cython: nonecheck=False

from aim.storage.encoding.encoding_native cimport *


cpdef inline bytes encode_int64_big_endian(int64 value):
    """Decode an 64-bit signed integer from a buffer
    encoded in big-endian
    """
    cdef char* buffer = [
        (value >> 56) & 0xff,
        (value >> 48) & 0xff,
        (value >> 40) & 0xff,
        (value >> 32) & 0xff,
        (value >> 24) & 0xff,
        (value >> 16) & 0xff,
        (value >> 8) & 0xff,
        (value >> 0) & 0xff
    ]
    return <bytes>buffer[:8]


cpdef inline bytes encode_int64(int64 value):
    """Encode an 64-bit signed integer into a buffer"""
    return (<unsigned char*>(&value))[:8]


cpdef inline int64 decode_int64_big_endian(const unsigned char* buf, int offset = 0) nogil:
    """Decode an 64-bit signed integer from a buffer
    encoded in big-endian
    """
    return (
        ((<int64>buf[0 + offset]) << 56) |
        ((<int64>buf[1 + offset]) << 48) |
        ((<int64>buf[2 + offset]) << 40) |
        ((<int64>buf[3 + offset]) << 32) |
        ((<int64>buf[4 + offset]) << 24) |
        ((<int64>buf[5 + offset]) << 16) |
        ((<int64>buf[6 + offset]) << 8) |
        ((<int64>buf[7 + offset]) << 0)
    )


cpdef inline int64 decode_int64(const unsigned char* buffer, int offset = 0) nogil:
    """Decode an 64-bit signed integer from a buffer"""
    return (<int64*>(buffer + offset))[0]


cpdef inline vector[pair[int64, int64]] split_path(
    const unsigned char* buffer,
    int64 length
) nogil:
    """Split the encoded path in order to decode into string and integer keys.
    The main goal of this functions is to quickly split the encoded path into
    smaller chunks in advance so the keys may be easily decoded.

    Args:
        buffer: binary-encoded path, e.g. for the path `('foo', 19, 'α-γ')`
                the encoded path is:
                `b'foo\\xfe\\xfe\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x13\\xfe\\xce\\xb1-\\xce\\xb3\\xfe'`
        length: the length of buffer

    Returns:
        An array of pairs denoting a key; If the key is integer then the value
        is encoded in `(value, 0)` otherwise the pair `(start, end)` denotes
        starting and ending offsets of key occurring in the provided buffer.
        For example, the result for the sample buffer will be:
          `[(0, 3),   # the segment from 0...3
            (19, 0),  # integer with value 19
            (14, 17)  # the segment 14...17
           ]`
    """
    cdef vector[pair[int64, int64]] path
    cdef int64 start = 0
    cdef int64 cursor = 0
    # We use `tmp_pair` to create a pair before pushing it into vector of chunks
    cdef pair[int64, int64] tmp_pair

    # Here we run over the buffer to detect the boundaries of key encodings.
    # The strategy is as follows:
    #  *  As we detect a new segment, we try to extend the boundary to right
    #     as long as it does not reach the path sentinel / separator symbol.
    #  *  When the cursor has reached a path separator, we store the resulting
    #     segment and forward the cursor right.
    #  *  If we detect a zero-length segment => we reached double-sentinel
    #     notation of integer keys in the path. In that case we decode an
    #     integer, store it in `(value, 0)` pair and forward the cursor right.
    while cursor < length:
        if buffer[cursor] != PATH_SENTINEL_CODE:
            # The current segment is `(start, cursor)` so extend it +1 to right
            cursor += 1
            continue

        if start < cursor:
            # The segment encodes a string key.
            # Store the segment boundaries
            tmp_pair.first = start
            tmp_pair.second = cursor
            path.push_back(tmp_pair)
        else:
            # The segment encodes a integer key
            # Store the integer in `(value, 0)` pair
            tmp_pair.first = decode_int64_big_endian(buffer + cursor + 1)
            tmp_pair.second = 0
            path.push_back(tmp_pair)
            # Move forward 1 byte over the integer key start sentinel,
            # then another 8 bytes for the 64-bit integer key
            cursor += 1 + 8

        # Making a new segment starting right after the path sentinel
        cursor += 1
        start = cursor

    return path


cpdef inline list decode_path(bytes buffer):
    """Decode the binary-encoded path from the given buffer.

    Args:
        buffer: binary-encoded path, e.g.
                `b'foo\\xfe\\xfe\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x13\\xfe\\xce\\xb1-\\xce\\xb3\\xfe'`

    Returns:
        The decoded path as a list of keys of type string or integer.
        For example, the corresponding path for the sample buffer will be:
        `['foo', 19, 'α-γ']`
    """
    cdef list path = []
    cdef vector[pair[int64, int64]] segments = split_path(buffer, len(buffer))
    for item in segments:
        if item.second == 0:
            # In this case the pair only encodes a single integer
            path.append(item.first)
        else:
            # The pair denotes start and end offsets of the string key
            # in the buffer encoded in utf-8
            path.append(buffer[item.first:item.second].decode('utf-8'))
    return path


cpdef inline bytes encode_double(double value):
    """Encode an double precision floating-point number into a buffer"""
    return (<unsigned char*>(&value))[:8]


cpdef inline double decode_double(const unsigned char* bytes, int offset = 0) nogil:
    """Decode an double precision floating-point number from a buffer"""
    return (<double*> (bytes + offset))[0]


cpdef inline bytes encode_utf_8_str(str value):
    """Encode an string into a buffer in utf-8"""
    return value.encode('utf-8')

cpdef inline str decode_utf_8_str(bytes buffer):
    """Decode an string from a buffer encoded in utf-8"""
    return buffer.decode('utf-8')



import json
from aim.web.api.dashboard_apps.models import ExploreState


def explore_state_response_serializer(es_object):
    if not isinstance(es_object, ExploreState):
        return None

    response_schema = {
        'id': es_object.uuid,
        'type': es_object.type,
        'updated_at': es_object.updated_at,
        'created_at': es_object.created_at,
        'state': json.loads(es_object.state)
    }

    return response_schema



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: remote_tracking.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='remote_tracking.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x15remote_tracking.proto\"J\n\x11\x45xceptionResponse\x12\x13\n\x0bmodule_name\x18\x01 \x01(\t\x12\x12\n\nclass_name\x18\x02 \x01(\t\x12\x0c\n\x04\x61rgs\x18\x03 \x01(\t\"2\n\x1c\x43lientResourceCleanupRequest\x12\x12\n\nclient_uri\x18\x01 \x01(\t\"\x9a\x01\n\x1d\x43lientResourceCleanupResponse\x12\x35\n\x06status\x18\x01 \x01(\x0e\x32%.ClientResourceCleanupResponse.Status\x12%\n\texception\x18\x02 \x01(\x0b\x32\x12.ExceptionResponse\"\x1b\n\x06Status\x12\x06\n\x02OK\x10\x00\x12\t\n\x05\x45RROR\x10\x01\"[\n\x0fResourceRequest\x12\x15\n\rresource_type\x18\x01 \x01(\t\x12\x0f\n\x07handler\x18\x02 \x01(\t\x12\x12\n\nclient_uri\x18\x03 \x01(\t\x12\x0c\n\x04\x61rgs\x18\x04 \x01(\x0c\"\x91\x01\n\x10ResourceResponse\x12\x0f\n\x07handler\x18\x01 \x01(\t\x12(\n\x06status\x18\x02 \x01(\x0e\x32\x18.ResourceResponse.Status\x12%\n\texception\x18\x03 \x01(\x0b\x32\x12.ExceptionResponse\"\x1b\n\x06Status\x12\x06\n\x02OK\x10\x00\x12\t\n\x05\x45RROR\x10\x01\"=\n\x16ReleaseResourceRequest\x12\x0f\n\x07handler\x18\x01 \x01(\t\x12\x12\n\nclient_uri\x18\x02 \x01(\t\"\x8e\x01\n\x17ReleaseResourceResponse\x12/\n\x06status\x18\x02 \x01(\x0e\x32\x1f.ReleaseResourceResponse.Status\x12%\n\texception\x18\x03 \x01(\x0b\x32\x12.ExceptionResponse\"\x1b\n\x06Status\x12\x06\n\x02OK\x10\x00\x12\t\n\x05\x45RROR\x10\x01\"Z\n\rRequestHeader\x12\x0f\n\x07version\x18\x01 \x01(\t\x12\x0f\n\x07handler\x18\x02 \x01(\t\x12\x12\n\nclient_uri\x18\x03 \x01(\t\x12\x13\n\x0bmethod_name\x18\x04 \x01(\t\"X\n\x12InstructionRequest\x12 \n\x06header\x18\x01 \x01(\x0b\x32\x0e.RequestHeaderH\x00\x12\x11\n\x07message\x18\x02 \x01(\x0cH\x00\x42\r\n\x0binstruction\"\x8d\x01\n\x0eResponseHeader\x12\x0f\n\x07version\x18\x01 \x01(\t\x12&\n\x06status\x18\x02 \x01(\x0e\x32\x16.ResponseHeader.Status\x12%\n\texception\x18\x03 \x01(\x0b\x32\x12.ExceptionResponse\"\x1b\n\x06Status\x12\x06\n\x02OK\x10\x00\x12\t\n\x05\x45RROR\x10\x01\"Z\n\x13InstructionResponse\x12!\n\x06header\x18\x01 \x01(\x0b\x32\x0f.ResponseHeaderH\x00\x12\x11\n\x07message\x18\x02 \x01(\x0cH\x00\x42\r\n\x0binstruction\"P\n\x18WriteInstructionsRequest\x12\x0f\n\x07version\x18\x01 \x01(\t\x12\x12\n\nclient_uri\x18\x02 \x01(\t\x12\x0f\n\x07message\x18\x03 \x01(\x0c\"\xa3\x01\n\x19WriteInstructionsResponse\x12\x0f\n\x07version\x18\x01 \x01(\t\x12\x31\n\x06status\x18\x02 \x01(\x0e\x32!.WriteInstructionsResponse.Status\x12%\n\texception\x18\x03 \x01(\x0b\x32\x12.ExceptionResponse\"\x1b\n\x06Status\x12\x06\n\x02OK\x10\x00\x12\t\n\x05\x45RROR\x10\x01\x32\x8d\x03\n\x15RemoteTrackingService\x12[\n\x18\x63leanup_client_resources\x12\x1d.ClientResourceCleanupRequest\x1a\x1e.ClientResourceCleanupResponse\"\x00\x12\x35\n\x0cget_resource\x12\x10.ResourceRequest\x1a\x11.ResourceResponse\"\x00\x12G\n\x10release_resource\x12\x17.ReleaseResourceRequest\x1a\x18.ReleaseResourceResponse\"\x00\x12\x42\n\x0frun_instruction\x12\x13.InstructionRequest\x1a\x14.InstructionResponse\"\x00(\x01\x30\x01\x12S\n\x16run_write_instructions\x12\x19.WriteInstructionsRequest\x1a\x1a.WriteInstructionsResponse\"\x00(\x01\x62\x06proto3'
)



_CLIENTRESOURCECLEANUPRESPONSE_STATUS = _descriptor.EnumDescriptor(
  name='Status',
  full_name='ClientResourceCleanupResponse.Status',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='OK', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ERROR', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=281,
  serialized_end=308,
)
_sym_db.RegisterEnumDescriptor(_CLIENTRESOURCECLEANUPRESPONSE_STATUS)

_RESOURCERESPONSE_STATUS = _descriptor.EnumDescriptor(
  name='Status',
  full_name='ResourceResponse.Status',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='OK', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ERROR', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=281,
  serialized_end=308,
)
_sym_db.RegisterEnumDescriptor(_RESOURCERESPONSE_STATUS)

_RELEASERESOURCERESPONSE_STATUS = _descriptor.EnumDescriptor(
  name='Status',
  full_name='ReleaseResourceResponse.Status',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='OK', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ERROR', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=281,
  serialized_end=308,
)
_sym_db.RegisterEnumDescriptor(_RELEASERESOURCERESPONSE_STATUS)

_RESPONSEHEADER_STATUS = _descriptor.EnumDescriptor(
  name='Status',
  full_name='ResponseHeader.Status',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='OK', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ERROR', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=281,
  serialized_end=308,
)
_sym_db.RegisterEnumDescriptor(_RESPONSEHEADER_STATUS)

_WRITEINSTRUCTIONSRESPONSE_STATUS = _descriptor.EnumDescriptor(
  name='Status',
  full_name='WriteInstructionsResponse.Status',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='OK', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ERROR', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=281,
  serialized_end=308,
)
_sym_db.RegisterEnumDescriptor(_WRITEINSTRUCTIONSRESPONSE_STATUS)


_EXCEPTIONRESPONSE = _descriptor.Descriptor(
  name='ExceptionResponse',
  full_name='ExceptionResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='module_name', full_name='ExceptionResponse.module_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='class_name', full_name='ExceptionResponse.class_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='args', full_name='ExceptionResponse.args', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=25,
  serialized_end=99,
)


_CLIENTRESOURCECLEANUPREQUEST = _descriptor.Descriptor(
  name='ClientResourceCleanupRequest',
  full_name='ClientResourceCleanupRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='client_uri', full_name='ClientResourceCleanupRequest.client_uri', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=101,
  serialized_end=151,
)


_CLIENTRESOURCECLEANUPRESPONSE = _descriptor.Descriptor(
  name='ClientResourceCleanupResponse',
  full_name='ClientResourceCleanupResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='ClientResourceCleanupResponse.status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exception', full_name='ClientResourceCleanupResponse.exception', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CLIENTRESOURCECLEANUPRESPONSE_STATUS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=154,
  serialized_end=308,
)


_RESOURCEREQUEST = _descriptor.Descriptor(
  name='ResourceRequest',
  full_name='ResourceRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='resource_type', full_name='ResourceRequest.resource_type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='handler', full_name='ResourceRequest.handler', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_uri', full_name='ResourceRequest.client_uri', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='args', full_name='ResourceRequest.args', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=310,
  serialized_end=401,
)


_RESOURCERESPONSE = _descriptor.Descriptor(
  name='ResourceResponse',
  full_name='ResourceResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='handler', full_name='ResourceResponse.handler', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='ResourceResponse.status', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exception', full_name='ResourceResponse.exception', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _RESOURCERESPONSE_STATUS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=404,
  serialized_end=549,
)


_RELEASERESOURCEREQUEST = _descriptor.Descriptor(
  name='ReleaseResourceRequest',
  full_name='ReleaseResourceRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='handler', full_name='ReleaseResourceRequest.handler', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_uri', full_name='ReleaseResourceRequest.client_uri', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=551,
  serialized_end=612,
)


_RELEASERESOURCERESPONSE = _descriptor.Descriptor(
  name='ReleaseResourceResponse',
  full_name='ReleaseResourceResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='ReleaseResourceResponse.status', index=0,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exception', full_name='ReleaseResourceResponse.exception', index=1,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _RELEASERESOURCERESPONSE_STATUS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=615,
  serialized_end=757,
)


_REQUESTHEADER = _descriptor.Descriptor(
  name='RequestHeader',
  full_name='RequestHeader',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='RequestHeader.version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='handler', full_name='RequestHeader.handler', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_uri', full_name='RequestHeader.client_uri', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='method_name', full_name='RequestHeader.method_name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=759,
  serialized_end=849,
)


_INSTRUCTIONREQUEST = _descriptor.Descriptor(
  name='InstructionRequest',
  full_name='InstructionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='InstructionRequest.header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='InstructionRequest.message', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='instruction', full_name='InstructionRequest.instruction',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=851,
  serialized_end=939,
)


_RESPONSEHEADER = _descriptor.Descriptor(
  name='ResponseHeader',
  full_name='ResponseHeader',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='ResponseHeader.version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='ResponseHeader.status', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exception', full_name='ResponseHeader.exception', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _RESPONSEHEADER_STATUS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=942,
  serialized_end=1083,
)


_INSTRUCTIONRESPONSE = _descriptor.Descriptor(
  name='InstructionResponse',
  full_name='InstructionResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='InstructionResponse.header', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='InstructionResponse.message', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='instruction', full_name='InstructionResponse.instruction',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=1085,
  serialized_end=1175,
)


_WRITEINSTRUCTIONSREQUEST = _descriptor.Descriptor(
  name='WriteInstructionsRequest',
  full_name='WriteInstructionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='WriteInstructionsRequest.version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_uri', full_name='WriteInstructionsRequest.client_uri', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='WriteInstructionsRequest.message', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1177,
  serialized_end=1257,
)


_WRITEINSTRUCTIONSRESPONSE = _descriptor.Descriptor(
  name='WriteInstructionsResponse',
  full_name='WriteInstructionsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='WriteInstructionsResponse.version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='WriteInstructionsResponse.status', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exception', full_name='WriteInstructionsResponse.exception', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _WRITEINSTRUCTIONSRESPONSE_STATUS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1260,
  serialized_end=1423,
)

_CLIENTRESOURCECLEANUPRESPONSE.fields_by_name['status'].enum_type = _CLIENTRESOURCECLEANUPRESPONSE_STATUS
_CLIENTRESOURCECLEANUPRESPONSE.fields_by_name['exception'].message_type = _EXCEPTIONRESPONSE
_CLIENTRESOURCECLEANUPRESPONSE_STATUS.containing_type = _CLIENTRESOURCECLEANUPRESPONSE
_RESOURCERESPONSE.fields_by_name['status'].enum_type = _RESOURCERESPONSE_STATUS
_RESOURCERESPONSE.fields_by_name['exception'].message_type = _EXCEPTIONRESPONSE
_RESOURCERESPONSE_STATUS.containing_type = _RESOURCERESPONSE
_RELEASERESOURCERESPONSE.fields_by_name['status'].enum_type = _RELEASERESOURCERESPONSE_STATUS
_RELEASERESOURCERESPONSE.fields_by_name['exception'].message_type = _EXCEPTIONRESPONSE
_RELEASERESOURCERESPONSE_STATUS.containing_type = _RELEASERESOURCERESPONSE
_INSTRUCTIONREQUEST.fields_by_name['header'].message_type = _REQUESTHEADER
_INSTRUCTIONREQUEST.oneofs_by_name['instruction'].fields.append(
  _INSTRUCTIONREQUEST.fields_by_name['header'])
_INSTRUCTIONREQUEST.fields_by_name['header'].containing_oneof = _INSTRUCTIONREQUEST.oneofs_by_name['instruction']
_INSTRUCTIONREQUEST.oneofs_by_name['instruction'].fields.append(
  _INSTRUCTIONREQUEST.fields_by_name['message'])
_INSTRUCTIONREQUEST.fields_by_name['message'].containing_oneof = _INSTRUCTIONREQUEST.oneofs_by_name['instruction']
_RESPONSEHEADER.fields_by_name['status'].enum_type = _RESPONSEHEADER_STATUS
_RESPONSEHEADER.fields_by_name['exception'].message_type = _EXCEPTIONRESPONSE
_RESPONSEHEADER_STATUS.containing_type = _RESPONSEHEADER
_INSTRUCTIONRESPONSE.fields_by_name['header'].message_type = _RESPONSEHEADER
_INSTRUCTIONRESPONSE.oneofs_by_name['instruction'].fields.append(
  _INSTRUCTIONRESPONSE.fields_by_name['header'])
_INSTRUCTIONRESPONSE.fields_by_name['header'].containing_oneof = _INSTRUCTIONRESPONSE.oneofs_by_name['instruction']
_INSTRUCTIONRESPONSE.oneofs_by_name['instruction'].fields.append(
  _INSTRUCTIONRESPONSE.fields_by_name['message'])
_INSTRUCTIONRESPONSE.fields_by_name['message'].containing_oneof = _INSTRUCTIONRESPONSE.oneofs_by_name['instruction']
_WRITEINSTRUCTIONSRESPONSE.fields_by_name['status'].enum_type = _WRITEINSTRUCTIONSRESPONSE_STATUS
_WRITEINSTRUCTIONSRESPONSE.fields_by_name['exception'].message_type = _EXCEPTIONRESPONSE
_WRITEINSTRUCTIONSRESPONSE_STATUS.containing_type = _WRITEINSTRUCTIONSRESPONSE
DESCRIPTOR.message_types_by_name['ExceptionResponse'] = _EXCEPTIONRESPONSE
DESCRIPTOR.message_types_by_name['ClientResourceCleanupRequest'] = _CLIENTRESOURCECLEANUPREQUEST
DESCRIPTOR.message_types_by_name['ClientResourceCleanupResponse'] = _CLIENTRESOURCECLEANUPRESPONSE
DESCRIPTOR.message_types_by_name['ResourceRequest'] = _RESOURCEREQUEST
DESCRIPTOR.message_types_by_name['ResourceResponse'] = _RESOURCERESPONSE
DESCRIPTOR.message_types_by_name['ReleaseResourceRequest'] = _RELEASERESOURCEREQUEST
DESCRIPTOR.message_types_by_name['ReleaseResourceResponse'] = _RELEASERESOURCERESPONSE
DESCRIPTOR.message_types_by_name['RequestHeader'] = _REQUESTHEADER
DESCRIPTOR.message_types_by_name['InstructionRequest'] = _INSTRUCTIONREQUEST
DESCRIPTOR.message_types_by_name['ResponseHeader'] = _RESPONSEHEADER
DESCRIPTOR.message_types_by_name['InstructionResponse'] = _INSTRUCTIONRESPONSE
DESCRIPTOR.message_types_by_name['WriteInstructionsRequest'] = _WRITEINSTRUCTIONSREQUEST
DESCRIPTOR.message_types_by_name['WriteInstructionsResponse'] = _WRITEINSTRUCTIONSRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ExceptionResponse = _reflection.GeneratedProtocolMessageType('ExceptionResponse', (_message.Message,), {
  'DESCRIPTOR' : _EXCEPTIONRESPONSE,
  '__module__' : 'remote_tracking_pb2'
  # @@protoc_insertion_point(class_scope:ExceptionResponse)
  })
_sym_db.RegisterMessage(ExceptionResponse)

ClientResourceCleanupRequest = _reflection.GeneratedProtocolMessageType('ClientResourceCleanupRequest', (_message.Message,), {
  'DESCRIPTOR' : _CLIENTRESOURCECLEANUPREQUEST,
  '__module__' : 'remote_tracking_pb2'
  # @@protoc_insertion_point(class_scope:ClientResourceCleanupRequest)
  })
_sym_db.RegisterMessage(ClientResourceCleanupRequest)

ClientResourceCleanupResponse = _reflection.GeneratedProtocolMessageType('ClientResourceCleanupResponse', (_message.Message,), {
  'DESCRIPTOR' : _CLIENTRESOURCECLEANUPRESPONSE,
  '__module__' : 'remote_tracking_pb2'
  # @@protoc_insertion_point(class_scope:ClientResourceCleanupResponse)
  })
_sym_db.RegisterMessage(ClientResourceCleanupResponse)

ResourceRequest = _reflection.GeneratedProtocolMessageType('ResourceRequest', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCEREQUEST,
  '__module__' : 'remote_tracking_pb2'
  # @@protoc_insertion_point(class_scope:ResourceRequest)
  })
_sym_db.RegisterMessage(ResourceRequest)

ResourceResponse = _reflection.GeneratedProtocolMessageType('ResourceResponse', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCERESPONSE,
  '__module__' : 'remote_tracking_pb2'
  # @@protoc_insertion_point(class_scope:ResourceResponse)
  })
_sym_db.RegisterMessage(ResourceResponse)

ReleaseResourceRequest = _reflection.GeneratedProtocolMessageType('ReleaseResourceRequest', (_message.Message,), {
  'DESCRIPTOR' : _RELEASERESOURCEREQUEST,
  '__module__' : 'remote_tracking_pb2'
  # @@protoc_insertion_point(class_scope:ReleaseResourceRequest)
  })
_sym_db.RegisterMessage(ReleaseResourceRequest)

ReleaseResourceResponse = _reflection.GeneratedProtocolMessageType('ReleaseResourceResponse', (_message.Message,), {
  'DESCRIPTOR' : _RELEASERESOURCERESPONSE,
  '__module__' : 'remote_tracking_pb2'
  # @@protoc_insertion_point(class_scope:ReleaseResourceResponse)
  })
_sym_db.RegisterMessage(ReleaseResourceResponse)

RequestHeader = _reflection.GeneratedProtocolMessageType('RequestHeader', (_message.Message,), {
  'DESCRIPTOR' : _REQUESTHEADER,
  '__module__' : 'remote_tracking_pb2'
  # @@protoc_insertion_point(class_scope:RequestHeader)
  })
_sym_db.RegisterMessage(RequestHeader)

InstructionRequest = _reflection.GeneratedProtocolMessageType('InstructionRequest', (_message.Message,), {
  'DESCRIPTOR' : _INSTRUCTIONREQUEST,
  '__module__' : 'remote_tracking_pb2'
  # @@protoc_insertion_point(class_scope:InstructionRequest)
  })
_sym_db.RegisterMessage(InstructionRequest)

ResponseHeader = _reflection.GeneratedProtocolMessageType('ResponseHeader', (_message.Message,), {
  'DESCRIPTOR' : _RESPONSEHEADER,
  '__module__' : 'remote_tracking_pb2'
  # @@protoc_insertion_point(class_scope:ResponseHeader)
  })
_sym_db.RegisterMessage(ResponseHeader)

InstructionResponse = _reflection.GeneratedProtocolMessageType('InstructionResponse', (_message.Message,), {
  'DESCRIPTOR' : _INSTRUCTIONRESPONSE,
  '__module__' : 'remote_tracking_pb2'
  # @@protoc_insertion_point(class_scope:InstructionResponse)
  })
_sym_db.RegisterMessage(InstructionResponse)

WriteInstructionsRequest = _reflection.GeneratedProtocolMessageType('WriteInstructionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _WRITEINSTRUCTIONSREQUEST,
  '__module__' : 'remote_tracking_pb2'
  # @@protoc_insertion_point(class_scope:WriteInstructionsRequest)
  })
_sym_db.RegisterMessage(WriteInstructionsRequest)

WriteInstructionsResponse = _reflection.GeneratedProtocolMessageType('WriteInstructionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _WRITEINSTRUCTIONSRESPONSE,
  '__module__' : 'remote_tracking_pb2'
  # @@protoc_insertion_point(class_scope:WriteInstructionsResponse)
  })
_sym_db.RegisterMessage(WriteInstructionsResponse)



_REMOTETRACKINGSERVICE = _descriptor.ServiceDescriptor(
  name='RemoteTrackingService',
  full_name='RemoteTrackingService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=1426,
  serialized_end=1823,
  methods=[
  _descriptor.MethodDescriptor(
    name='cleanup_client_resources',
    full_name='RemoteTrackingService.cleanup_client_resources',
    index=0,
    containing_service=None,
    input_type=_CLIENTRESOURCECLEANUPREQUEST,
    output_type=_CLIENTRESOURCECLEANUPRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='get_resource',
    full_name='RemoteTrackingService.get_resource',
    index=1,
    containing_service=None,
    input_type=_RESOURCEREQUEST,
    output_type=_RESOURCERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='release_resource',
    full_name='RemoteTrackingService.release_resource',
    index=2,
    containing_service=None,
    input_type=_RELEASERESOURCEREQUEST,
    output_type=_RELEASERESOURCERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='run_instruction',
    full_name='RemoteTrackingService.run_instruction',
    index=3,
    containing_service=None,
    input_type=_INSTRUCTIONREQUEST,
    output_type=_INSTRUCTIONRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='run_write_instructions',
    full_name='RemoteTrackingService.run_write_instructions',
    index=4,
    containing_service=None,
    input_type=_WRITEINSTRUCTIONSREQUEST,
    output_type=_WRITEINSTRUCTIONSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_REMOTETRACKINGSERVICE)

DESCRIPTOR.services_by_name['RemoteTrackingService'] = _REMOTETRACKINGSERVICE

# @@protoc_insertion_point(module_scope)



import os
from pathlib import Path

from fastapi import HTTPException, Request
from fastapi.responses import FileResponse, HTMLResponse

from aim.web.api.utils import APIRouter  # wrapper for fastapi.APIRouter
from aim.web.configs import AIM_UI_BASE_PATH

statics_router = APIRouter()


@statics_router.get('/static-files/{path:path}/')
async def serve_static_files(path):
    import aim_ui

    static_files_root = os.path.join(os.path.dirname(aim_ui.__file__), 'build')
    static_file_name = '/'.join((static_files_root, path))

    # check if path is leading inside ui/build directory
    if not Path(static_files_root).resolve() in Path(static_file_name).resolve().parents:
        raise HTTPException(status_code=404)

    compressed_file_name = '{}.gz'.format(static_file_name)
    if os.path.exists(compressed_file_name):
        return FileResponse(compressed_file_name, headers={'Content-Encoding': 'gzip'})
    return FileResponse(static_file_name)


# do not change the placement of this method
# as it also serves as a fallback for wrong url routes
@statics_router.get('/{path:path}/', response_class=HTMLResponse)
async def serve_index_html(request: Request):
    import aim_ui
    from jinja2 import Environment, FileSystemLoader

    template_files_dir = os.path.join(os.path.dirname(aim_ui.__file__), 'build')
    env = Environment(
        loader=FileSystemLoader(template_files_dir),
        autoescape=True
    )
    template = env.get_template('index-template.html')
    base_path = os.environ.get(AIM_UI_BASE_PATH, '')
    return template.render(base_path=base_path)



import logging

from aim.ext.notifier.logging_notifier import LoggingNotifier
from aim.ext.notifier.slack_notifier import SlackNotifier
from aim.ext.notifier.workplace_notifier import WorkplaceNotifier
from aim.ext.notifier.notifier import Notifier

logger = logging.getLogger(__name__)


class NotifierBuilder(object):
    _factories = {
        'logger': LoggingNotifier,
        'workplace': WorkplaceNotifier,
        'slack': SlackNotifier
    }

    def __init__(self, config: dict):
        self.config = config

    def build(self) -> Notifier:
        notifier = Notifier()
        for sub_config in self.config.values():
            notif_type = sub_config['type']
            notif_id = sub_config['id']
            args = sub_config['arguments']
            if sub_config['status'] != 'enabled':
                continue
            try:
                notif_cls = self._factories[notif_type]
            except KeyError:
                logger.warning(f'Unknown notifier type {notif_type}. Skipping.')
            else:
                sub_notifier = notif_cls(notif_id, args)
                notifier.add(sub_notifier)
        return notifier



def get_inst_type_str(inst):
    """
    Get instance type and class type full names
    """
    obj_name = obj_module = obj_cls_name = obj_cls_module = ''

    if hasattr(inst, '__name__'):
        obj_name = inst.__name__
    if hasattr(inst, '__module__'):
        obj_module = inst.__module__
    if hasattr(inst, '__class__'):
        if hasattr(inst.__class__, '__name__'):
            obj_cls_name = inst.__class__.__name__
        if hasattr(inst.__class__, '__module__'):
            obj_cls_module = inst.__class__.__module__

    obj_full = '{}.{}'.format(obj_name, obj_module)
    obj_cls_full = '{}.{}'.format(obj_cls_name, obj_cls_module)

    return obj_full, obj_cls_full


def get_inst_base_types(inst):
    """
    Get instance and it's base classes types
    """
    bases_types = []
    for b in inst.__class__.__bases__:
        b_type, b_cls_type = get_inst_type_str(b)
        bases_types.append(b_type)
        bases_types.append(b_cls_type)
    return bases_types


def inst_has_typename(inst, types):
    """
    Return `True` if the instance is created from class
    which has base that matches passed `types`
    """
    inst_type, inst_cls_type = get_inst_type_str(inst)
    inst_types = [inst_type, inst_cls_type] + get_inst_base_types(inst)

    for i in inst_types:
        found = True
        for t in types:
            if i.find(t) == -1:
                found = False
                break
        if found:
            return True

    return False


def is_pytorch_tensor(inst):
    """
    Check whether `inst` is instance of pytorch tensor
    """
    return inst_has_typename(inst, ['torch', 'Tensor'])


def is_tf_tensor(inst):
    return inst_has_typename(inst, ['tensorflow', 'Tensor'])


def is_jax_device_array(inst):
    """
    Check whether `inst` is instance of jax device array
    """
    return inst_has_typename(inst, ['jaxlib', 'xla_extension', 'DeviceArray'])


def is_numpy_array(inst):
    """
    Check whether `inst` is instance of numpy array
    """
    return inst_has_typename(inst, ['numpy', 'ndarray'])


def is_numpy_number(inst):
    """
    Check whether `inst` is numpy number
    """

    return inst_has_typename(inst, ['numpy'])


def is_py_number(value):
    return isinstance(value, (int, float))


def is_number(value):
    """
    Checks if the given value is a number
    """
    if is_py_number(value):
        return True

    if is_numpy_array(value):
        return True

    if is_numpy_number(value):
        return True

    if is_jax_device_array(value):
        return True

    if is_pytorch_tensor(value):
        return True

    if is_tf_tensor(value):
        return True

    return False


def convert_to_py_number(value) -> object:
    """
    Converts numpy objects or tensors to python number types
    """
    if isinstance(value, int):
        return int(value)

    if isinstance(value, float):
        return float(value)

    if is_numpy_array(value):
        return value.item()

    if is_numpy_number(value):
        return value.item()

    if is_jax_device_array(value):
        return value.item()

    if is_pytorch_tensor(value):
        return value.item()

    if is_tf_tensor(value):
        return value.numpy().item()

    raise ValueError('not a number')



import json

from fastapi import Depends, HTTPException
from aim.web.api.utils import APIRouter  # wrapper for fastapi.APIRouter
from sqlalchemy.orm import Session

from aim.web.api.dashboard_apps.models import ExploreState
from aim.web.api.dashboard_apps.pydantic_models import (
    ExploreStateCreateIn,
    ExploreStateUpdateIn,
    ExploreStateGetOut,
    ExploreStateListOut
)
from aim.web.api.dashboard_apps.serializers import explore_state_response_serializer
from aim.web.api.db import get_session

dashboard_apps_router = APIRouter()


@dashboard_apps_router.get('/', response_model=ExploreStateListOut)
async def dashboard_apps_list_api(session: Session = Depends(get_session)):
    explore_states = session.query(ExploreState).filter(ExploreState.is_archived == False)  # noqa
    result = []
    for es in explore_states:
        result.append(explore_state_response_serializer(es))
    return result


@dashboard_apps_router.post('/', status_code=201, response_model=ExploreStateGetOut)
async def dashboard_apps_create_api(explore_state_in: ExploreStateCreateIn, session: Session = Depends(get_session)):
    explore_state = ExploreState()
    explore_state.type = explore_state_in.type
    explore_state.state = json.dumps(explore_state_in.state)

    session.add(explore_state)
    session.commit()

    return explore_state_response_serializer(explore_state)


@dashboard_apps_router.get('/{app_id}/', response_model=ExploreStateGetOut)
async def dashboard_apps_get_api(app_id: str, session: Session = Depends(get_session)):
    explore_state = session.query(ExploreState) \
        .filter(ExploreState.uuid == app_id, ExploreState.is_archived == False) \
        .first()
    if not explore_state:
        raise HTTPException(status_code=404)

    return explore_state_response_serializer(explore_state)


@dashboard_apps_router.put('/{app_id}/', response_model=ExploreStateGetOut)
async def dashboard_apps_put_api(app_id: str, explore_state_in: ExploreStateUpdateIn,
                                 session: Session = Depends(get_session)):
    explore_state = session.query(ExploreState) \
        .filter(ExploreState.uuid == app_id, ExploreState.is_archived == False) \
        .first() # noqa
    if not explore_state:
        raise HTTPException(status_code=404)

    if explore_state_in.type:
        explore_state.type = explore_state_in.type
    if explore_state_in.state:
        explore_state.state = json.dumps(explore_state_in.state)

    session.commit()

    return explore_state_response_serializer(explore_state)


@dashboard_apps_router.delete('/{app_id}/')
async def dashboard_apps_delete_api(app_id: str, session: Session = Depends(get_session)):
    explore_state = session.query(ExploreState) \
        .filter(ExploreState.uuid == app_id, ExploreState.is_archived == False) \
        .first()  # noqa
    if not explore_state:
        raise HTTPException(status_code=404)

    explore_state.is_archived = True
    session.commit()



import uuid
import sqlalchemy as sa

from aim.web.api.db import Base
from aim.web.api.utils import datetime_now


class Dashboard(Base):
    __tablename__ = 'dashboards'
    uuid = sa.Column(sa.Text, primary_key=True)
    name = sa.Column(sa.Text)
    description = sa.Column(sa.Text)

    created_at = sa.Column(sa.DateTime, default=datetime_now)
    updated_at = sa.Column(sa.DateTime, default=datetime_now, onupdate=datetime_now)
    is_archived = sa.Column(sa.Boolean, default=False)

    def __init__(self, name, description):
        self.uuid = str(uuid.uuid1())
        self.name = name
        self.description = description
        self.is_archived = False



# distutils: language = c++
# cython: language_level = 3

cdef class ArrayView:
    """Array of homogeneous elements with sparse indices.
    Interface for working with array as a non-sparse array is available for cases
    when index values are not important.
    """
    pass



import time
import numpy as np
from mxnet.gluon.contrib.estimator.utils import _check_metrics
from mxnet.gluon.contrib.estimator import TrainBegin, TrainEnd, EpochBegin, EpochEnd, BatchBegin, BatchEnd, Estimator
from typing import Optional, Union, Any, List
from aim.sdk.run import Run
from aim.ext.resource.configs import DEFAULT_SYSTEM_TRACKING_INT


class AimLoggingHandler(TrainBegin, TrainEnd, EpochBegin, EpochEnd, BatchBegin, BatchEnd):
    """Aim wrapper on top of the Mxnet Basic Logging Handler that applies to every Gluon estimator by default.
    :py:class:`AimLoggingHandler` logs hyper-parameters, training statistics,
    and other useful information during training
    Parameters
    ----------
    log_interval: int or str, default 'epoch'
        Logging interval during training.
        log_interval='epoch': display metrics every epoch
        log_interval=integer k: display metrics every interval of k batches
    metrics : list of EvalMetrics
        Metrics to be logged, logged at batch end, epoch end, train end.
    priority : scalar, default np.Inf
        Priority level of the AimLoggingHandler. Priority level is sorted in
        ascending order. The lower the number is, the higher priority level the
        handler is.
    """

    def __init__(self, log_interval: Union[int, str] = 'epoch',
                 repo: Optional[str] = None,
                 experiment_name: Optional[str] = None,
                 system_tracking_interval: Optional[int] = DEFAULT_SYSTEM_TRACKING_INT,
                 log_system_params: bool = True,
                 metrics: List[Any] = None,
                 priority=np.Inf,):
        super().__init__()
        if not isinstance(log_interval, int) and log_interval != 'epoch':
            raise ValueError("log_interval must be either an integer or string 'epoch'")

        self.metrics = _check_metrics(metrics)
        self.batch_index = 0
        self.current_epoch = 0
        self.processed_samples = 0
        # logging handler need to be called at last to make sure all states are updated
        # it will also shut down logging at train end
        self.priority = priority
        self.log_interval = log_interval
        self.log_interval_time = 0

        self.repo = repo
        self.experiment_name = experiment_name
        self.system_tracking_interval = system_tracking_interval
        self.log_system_params = log_system_params
        self._run = None
        self._run_hash = None

    def train_begin(self, estimator: Optional[Estimator], *args, **kwargs):
        self.train_start = time.time()
        trainer = estimator.trainer
        optimizer = trainer.optimizer.__class__.__name__
        lr = trainer.learning_rate

        estimator.logger.info("Training begin: using optimizer %s "
                              "with current learning rate %.4f ",
                              optimizer, lr)
        if estimator.max_epoch:
            estimator.logger.info("Train for %d epochs.", estimator.max_epoch)
        else:
            estimator.logger.info("Train for %d batches.", estimator.max_batch)
        # reset all counters
        self.current_epoch = 0
        self.batch_index = 0
        self.processed_samples = 0
        self.log_interval_time = 0

        params = {
            "arch": estimator.net.name,
            "loss": estimator.loss.name,
            "optimizer": optimizer,
            "lr": lr,
            "max_epoch": estimator.max_epoch,
            "max_batch": estimator.max_batch
        }

        self.setup(estimator, params)

    def train_end(self, estimator: Optional[Estimator], *args, **kwargs):
        train_time = time.time() - self.train_start
        msg = 'Train finished using total %ds with %d epochs. ' % (train_time, self.current_epoch)
        # log every result in train stats including train/validation loss & metrics
        for metric in self.metrics:
            name, value = metric.get()
            msg += '%s: %.4f, ' % (name, value)
        estimator.logger.info(msg.rstrip(', '))

    def epoch_begin(self, estimator: Optional[Estimator], *args, **kwargs):
        if isinstance(self.log_interval, int) or self.log_interval == 'epoch':
            is_training = False
            for metric in self.metrics:
                if 'training' in metric.name:
                    is_training = True
            self.epoch_start = time.time()
            if is_training:
                estimator.logger.info("[Epoch %d] Begin, current learning rate: %.4f",
                                      self.current_epoch, estimator.trainer.learning_rate)
            else:
                estimator.logger.info("Validation Begin")

    def epoch_end(self, estimator: Optional[Estimator], *args, **kwargs):
        if isinstance(self.log_interval, int) or self.log_interval == 'epoch':
            epoch_time = time.time() - self.epoch_start
            msg = '[Epoch %d] Finished in %.3fs, ' % (self.current_epoch, epoch_time)
            for metric in self.metrics:
                name, value = metric.get()
                msg += '%s: %.4f, ' % (name, value)

                context_name, metric_name = name.split(" ")
                context = {'subset': context_name}
                self._run.track(value, metric_name, step=self.batch_index, context=context)
            estimator.logger.info(msg.rstrip(', '))
        self.current_epoch += 1
        self.batch_index = 0

    def batch_begin(self, estimator: Optional[Estimator], *args, **kwargs):
        if isinstance(self.log_interval, int):
            self.batch_start = time.time()

    def batch_end(self, estimator: Optional[Estimator], *args, **kwargs):
        if isinstance(self.log_interval, int):
            batch_time = time.time() - self.batch_start
            msg = '[Epoch %d][Batch %d]' % (self.current_epoch, self.batch_index)
            self.processed_samples += kwargs['batch'][0].shape[0]
            msg += '[Samples %s] ' % (self.processed_samples)
            self.log_interval_time += batch_time
            if self.batch_index % self.log_interval == 0:
                msg += 'time/interval: %.3fs ' % self.log_interval_time
                self.log_interval_time = 0
                for metric in self.metrics:
                    # only log current training loss & metric after each interval
                    name, value = metric.get()
                    msg += '%s: %.4f, ' % (name, value)

                    context_name, metric_name = name.split(" ")
                    context = {'subset': context_name}
                    self._run.track(value, metric_name, step=self.batch_index, context=context)
                estimator.logger.info(msg.rstrip(', '))
        self.batch_index += 1

    @property
    def experiment(self) -> Run:
        if not self._run:
            self.setup()
        return self._run

    def setup(self, estimator: Optional[Estimator] = None, args=None):
        if not self._run:
            if self._run_hash:
                self._run = Run(
                    self._run_hash,
                    repo=self.repo,
                    system_tracking_interval=self.system_tracking_interval,
                    log_system_params=self.log_system_params,
                )
            else:
                self._run = Run(
                    repo=self.repo,
                    experiment=self.experiment_name,
                    system_tracking_interval=self.system_tracking_interval,
                    log_system_params=self.log_system_params,
                )
                self._run_hash = self._run.hash

        # Log config parameters
        if args:
            try:
                for key in args:
                    self._run.set(key, args[key], strict=False)
            except Exception as e:
                estimator.logger.warning(f'Aim could not log config parameters -> {e}')

    def __del__(self):
        if self._run and self._run.active:
            self._run.close()



from parameterized import parameterized

from aim import Repo

from performance_tests.base import StorageTestBase
from performance_tests.utils import get_baseline, write_baseline
from performance_tests.storage.utils import random_access_metric_values


class TestRandomAccess(StorageTestBase):
    @parameterized.expand({0: 50, 1: 250, 2: 500}.items())
    def test_random_access(self, test_key, density):
        test_name = f'test_random_access_{test_key}'
        repo = Repo.default_repo()
        query = 'metric.name == "metric 0"'
        execution_time = random_access_metric_values(repo, query, density)
        baseline = get_baseline(test_name)
        if baseline:
            self.assertInRange(execution_time, baseline)
        else:
            write_baseline(test_name, execution_time)



import os
import uuid

from aim.ext.transport.config import AIM_SERVER_MOUNTED_REPO_PATH
from aim.ext.transport.message_utils import unpack_args

from aim.sdk import Repo
from aim.sdk.reporter import RunStatusReporter, ScheduledStatusReporter
from aim.ext.cleanup import AutoClean
from aim.storage.treeutils import decode_tree


class ResourceRefAutoClean(AutoClean['ResourceRef']):
    @staticmethod
    def noop(res: object):
        return

    def __init__(self, instance: 'ResourceRef'):
        super().__init__(instance)
        self._finalizer_func = instance._finalizer_func
        self._resource = instance._resource

    def _close(self):
        self._finalizer_func(self._resource)


class ResourceRef:
    def __init__(self, res_obj, finalizer_func=ResourceRefAutoClean.noop):
        self._resource = res_obj
        self._finalizer_func = finalizer_func
        self._auto_clean = ResourceRefAutoClean(self)

    @property
    def ref(self):
        return self._resource


def get_handler():
    return str(uuid.uuid4())


def get_tree(args: bytes):
    repo_path = os.environ.get(AIM_SERVER_MOUNTED_REPO_PATH)
    if repo_path:
        repo = Repo.from_path(repo_path)
    else:
        repo = Repo.default_repo()
    kwargs = decode_tree(unpack_args(args))
    name = kwargs['name']
    sub = kwargs['sub']
    read_only = kwargs['read_only']
    from_union = kwargs['from_union']
    index = kwargs['index']
    timeout = kwargs['timeout']
    if index:
        return ResourceRef(repo._get_index_tree(name, timeout))
    else:
        return ResourceRef(repo.request_tree(name, sub, read_only=read_only, from_union=from_union))


def get_structured_run(args: bytes):
    repo_path = os.environ.get(AIM_SERVER_MOUNTED_REPO_PATH)
    if repo_path:
        repo = Repo.from_path(repo_path)
    else:
        repo = Repo.default_repo()

    kwargs = decode_tree(unpack_args(args))
    hash_ = kwargs['hash_']
    read_only = kwargs['read_only']

    return ResourceRef(repo.request_props(hash_, read_only))


def get_repo(args: bytes):
    repo_path = os.environ.get(AIM_SERVER_MOUNTED_REPO_PATH)
    if repo_path:
        repo = Repo.from_path(repo_path)
    else:
        repo = Repo.default_repo()
    return ResourceRef(repo)


def get_lock(args: bytes):
    repo_path = os.environ.get(AIM_SERVER_MOUNTED_REPO_PATH)
    if repo_path:
        repo = Repo.from_path(repo_path)
    else:
        repo = Repo.default_repo()
    kwargs = decode_tree(unpack_args(args))
    run_hash = kwargs['run_hash']
    # TODO Do we need to import SFRunLock here?
    from aim.sdk.lock_manager import SFRunLock
    return ResourceRef(repo.request_run_lock(run_hash), SFRunLock.release)


def get_run_heartbeat(args: bytes):
    repo_path = os.environ.get(AIM_SERVER_MOUNTED_REPO_PATH)
    if repo_path:
        repo = Repo.from_path(repo_path)
    else:
        repo = Repo.default_repo()
    kwargs = decode_tree(unpack_args(args))
    run_hash = kwargs['run_hash']
    return ResourceRef(ScheduledStatusReporter(RunStatusReporter(run_hash, repo.path)), ScheduledStatusReporter.stop)



import click
import os

from aim.sdk.repo import Repo
from aim.sdk.utils import clean_repo_path


@click.command()
@click.option('--repo', required=False, type=click.Path(exists=True,
                                                        file_okay=False,
                                                        dir_okay=True,
                                                        writable=True))
def init(repo):
    """
    Initializes new repository in the --repo directory.
    Initializes new repository in the current working directory if --repo argument is not provided:
     - Creates .aim directory & runs upgrades for structured DB
    """
    repo_path = clean_repo_path(repo) or os.getcwd()
    re_init = False
    if Repo.exists(repo_path):
        re_init = click.confirm('Aim repository is already initialized. '
                                'Do you want to re-initialize to empty Aim repository?')
        if not re_init:
            return
        # Clear old repo
        Repo.rm(repo_path)

    repo = Repo.from_path(repo_path, init=True)
    if re_init:
        click.echo(
            'Re-initialized empty Aim repository at {}'.format(repo.root_path))
    else:
        click.echo('Initialized a new Aim repository at {}'.format(repo.root_path))



from pydantic import BaseModel
from typing import Dict, List, Tuple, Optional, Union
from uuid import UUID


# response models
class EncodedNumpyArray(BaseModel):
    type: str = 'numpy'
    shape: int = 0
    dtype: str = 'float64'
    blob: bytes = ''


class TraceBase(BaseModel):
    context: dict
    name: str


class TraceOverview(TraceBase):
    last_value: float = 0.1


class TraceBaseView(TraceBase):
    iters: List[int]


class MetricsBaseView(TraceBaseView):
    values: List[float]


RunMetricsBatchApiOut = List[MetricsBaseView]


class TraceAlignedView(TraceBase):
    x_axis_values: Optional[EncodedNumpyArray] = None
    x_axis_iters: Optional[EncodedNumpyArray] = None


RunMetricCustomAlignApiOut = Dict[str, List[TraceAlignedView]]


class TraceFullView(TraceAlignedView):
    slice: Tuple[int, int, int]
    values: Optional[EncodedNumpyArray] = None
    iters: Optional[EncodedNumpyArray] = None
    epochs: Optional[EncodedNumpyArray] = None
    timestamps: Optional[EncodedNumpyArray] = None


class PropsView(BaseModel):
    class Tag(BaseModel):
        id: UUID
        name: str
        color: Optional[str] = None
        description: Optional[str] = None

    class Experiment(BaseModel):
        id: UUID
        name: str
        description: Optional[str] = None

    name: Optional[str] = None
    description: Optional[str] = None
    experiment: Optional[Experiment] = None
    tags: Optional[List[Tag]] = []
    creation_time: float
    end_time: Optional[float]
    archived: bool
    active: bool


class MetricSearchRunView(BaseModel):
    params: dict
    traces: List[TraceFullView]
    props: PropsView


class RunInfoOut(BaseModel):
    params: dict
    traces: Dict[str, List[TraceOverview]]
    props: PropsView


RunMetricSearchApiOut = Dict[str, MetricSearchRunView]


class RunSearchRunView(BaseModel):
    params: Optional[dict]
    traces: Optional[List[TraceOverview]]
    props: PropsView


RunSearchApiOut = Dict[str, RunSearchRunView]


class RunActiveOut(BaseModel):
    traces: Dict[str, List[TraceOverview]]
    props: PropsView


# request models
class AlignedTraceIn(BaseModel):
    context: dict
    name: str
    slice: Tuple[int, int, int]


class AlignedRunIn(BaseModel):
    run_id: str
    traces: List[AlignedTraceIn]


class MetricAlignApiIn(BaseModel):
    align_by: str
    runs: List[AlignedRunIn]


RunTracesBatchApiIn = List[TraceBase]


# structured run models
class StructuredRunUpdateIn(BaseModel):
    name: Optional[str] = None
    description: Optional[str] = None
    archived: Optional[bool] = None
    experiment: Optional[str] = None


class StructuredRunUpdateOut(BaseModel):
    id: str
    status: str = 'OK'


class StructuredRunsArchivedOut(BaseModel):
    status: str = 'OK'


class StructuredRunAddTagIn(BaseModel):
    tag_name: str


class StructuredRunAddTagOut(BaseModel):
    id: str
    tag_id: UUID
    status: str = 'OK'


class StructuredRunRemoveTagOut(BaseModel):
    id: str
    removed: bool
    status: str = 'OK'


class QuerySyntaxErrorOut(BaseModel):
    class SE(BaseModel):
        name: str
        statement: str
        line: int
        offset: int

    detail: SE


URIBatchIn = List[str]
RunsBatchIn = List[str]


# Custom object Models "Fully Generic"

class BaseRangeInfo(BaseModel):
    record_range_used: Tuple[int, int]
    record_range_total: Tuple[int, int]
    index_range_used: Optional[Tuple[int, int]]
    index_range_total: Optional[Tuple[int, int]]


class ObjectSequenceBaseView(BaseRangeInfo, TraceBaseView):
    values: list


class ObjectSequenceFullView(TraceBaseView):
    values: list
    iters: List[int]
    epochs: List[int]
    timestamps: List[float]


class ObjectSearchRunView(BaseModel):
    params: dict
    traces: List[ObjectSequenceFullView]
    ranges: BaseRangeInfo
    props: PropsView


# Custom objects
class ImageInfo(BaseModel):
    caption: str
    width: int
    height: int
    blob_uri: str
    index: int


class TextInfo(BaseModel):
    data: str
    index: int


class AudioInfo(BaseModel):
    caption: str
    blob_uri: str
    index: int


class DistributionInfo(BaseModel):
    data: EncodedNumpyArray
    bin_count: int
    range: Tuple[Union[int, float], Union[int, float]]


class FigureInfo(BaseModel):
    blob_uri: str


class NoteIn(BaseModel):
    content: str


ImageList = List[ImageInfo]
TextList = List[TextInfo]
AudioList = List[AudioInfo]
